import hashlib
import json
import os
import re
import shutil
import sys
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, List, Optional, Set, Union
from urllib.parse import urlparse

import pytest

from aqt.exceptions import (
    AqtException,
    ArchiveConnectionError,
    ArchiveDownloadError,
    ArchiveListError,
    CliInputError,
    EmptyMetadata,
)
from aqt.helper import Settings
from aqt.installer import Cli
from aqt.metadata import (
    ArchiveId,
    MetadataFactory,
    QtRepoProperty,
    SimpleSpec,
    ToolData,
    Version,
    Versions,
    show_list,
    suggested_follow_up,
)

ModulesQuery = MetadataFactory.ModulesQuery

Settings.load_settings()


@pytest.mark.parametrize(
    "arch, version, expect",
    (
        ("wasm_32", Version("5.13.1"), "wasm"),
        ("mingw", Version("5.13.1"), ""),
        ("android_fake", Version("5.13.1"), ""),
        ("android_x86", Version("5.13.1"), ""),
        ("android_x86", Version("6.13.1"), "x86"),
        ("android_x86", Version("6.0.0"), "x86"),
    ),
)
def test_list_extension_for_arch(arch: str, version: Version, expect: str):
    ext = QtRepoProperty.extension_for_arch(arch, version >= Version("6.0.0"))
    assert ext == expect


@pytest.mark.parametrize(
    "arch, expect",
    (
        ("wasm_32", ["wasm"]),
        ("mingw", [""]),
        ("android_fake", [""]),
        ("android", [""]),
        ("android_x86", ["", "x86"]),
    ),
)
def test_list_possible_extension_for_arch(arch: str, expect: List[str]):
    exts = QtRepoProperty.possible_extensions_for_arch(arch)
    assert set(exts) == set(expect)


@pytest.mark.parametrize(
    "init_data, expect_str, expect_fmt, expect_flat, expect_last, expect_bool",
    (
        (
            [
                (1, [Version("1.1.1"), Version("1.1.2")]),
                (2, [Version("1.2.1"), Version("1.2.2")]),
            ],
            "[[Version('1.1.1'), Version('1.1.2')], [Version('1.2.1'), Version('1.2.2')]]",
            "1.1.1 1.1.2\n1.2.1 1.2.2",
            [Version("1.1.1"), Version("1.1.2"), Version("1.2.1"), Version("1.2.2")],
            Version("1.2.2"),
            True,
        ),
        (
            [],
            "[]",
            "",
            [],
            None,
            False,
        ),
        (
            Version("1.2.3"),
            "[[Version('1.2.3')]]",
            "1.2.3",
            [Version("1.2.3")],
            Version("1.2.3"),
            True,
        ),
    ),
)
def test_versions(init_data, expect_str, expect_fmt, expect_flat, expect_last, expect_bool):
    versions = Versions(init_data)
    assert str(versions) == expect_str
    assert format(versions) == expect_fmt
    assert format(versions, "s") == expect_str
    assert versions.flattened() == expect_flat
    assert versions.latest() == expect_last
    assert bool(versions) == expect_bool

    with pytest.raises(TypeError) as pytest_wrapped_e:
        format(versions, "x")
    assert pytest_wrapped_e.type == TypeError


@pytest.fixture
def spec_regex():
    return re.compile(r"^(\d+\.\d+)")


@pytest.mark.parametrize(
    "os_name,target,in_file,expect_out_file",
    [
        ("windows", "android", "windows-android.html", "windows-android-expect.json"),
        ("windows", "desktop", "windows-desktop.html", "windows-desktop-expect.json"),
        ("windows", "winrt", "windows-winrt.html", "windows-winrt-expect.json"),
        ("linux", "android", "linux-android.html", "linux-android-expect.json"),
        ("linux", "desktop", "linux-desktop.html", "linux-desktop-expect.json"),
        ("mac", "android", "mac-android.html", "mac-android-expect.json"),
        ("mac", "desktop", "mac-desktop.html", "mac-desktop-expect.json"),
        ("mac", "ios", "mac-ios.html", "mac-ios-expect.json"),
        ("windows", "android", "mirror-pre-a.html", "mirror-expect.json"),
        ("windows", "android", "mirror-table-before-pre-a.html", "mirror-expect.json"),
        ("windows", "android", "mirror-first-td.html", "mirror-expect.json"),
        ("windows", "android", "mirror-tag-in-a.html", "mirror-expect.json"),
    ],
)
def test_list_versions_tools(monkeypatch, spec_regex, os_name, target, in_file, expect_out_file):
    _html = (Path(__file__).parent / "data" / in_file).read_text("utf-8")
    monkeypatch.setattr(MetadataFactory, "fetch_http", lambda *args, **kwargs: _html)

    expected = json.loads((Path(__file__).parent / "data" / expect_out_file).read_text("utf-8"))

    # Test 'aqt list-tool'
    tools = MetadataFactory(ArchiveId("tools", os_name, target)).getList()
    assert tools == expected["tools"]

    # Test 'aqt list-qt'
    expected_output = expected["qt"]["qt"]
    archive_id = ArchiveId("qt", os_name, target)
    all_versions = MetadataFactory(archive_id).getList()
    assert f"{all_versions}" == "\n".join(expected_output)

    # Filter for the latest version only
    latest_ver = MetadataFactory(archive_id, is_latest_version=True).getList()
    assert f"{latest_ver}" == expected_output[-1].split(" ")[-1]

    for row in expected_output:
        spec = SimpleSpec(spec_regex.search(row).group(1))

        # Find the latest version for a particular spec
        latest_ver_for_spec = MetadataFactory(
            archive_id,
            spec=spec,
            is_latest_version=True,
        ).getList()
        assert f"{latest_ver_for_spec}" == row.split(" ")[-1]

        # Find all versions for a particular spec
        all_ver_for_spec = MetadataFactory(
            archive_id,
            spec=spec,
        ).getList()
        assert f"{all_ver_for_spec}" == row


@pytest.mark.parametrize(
    "version,extension,in_file,expect_out_file",
    [
        ("5.14.0", "", "windows-5140-update.xml", "windows-5140-expect.json"),
        ("5.15.0", "", "windows-5150-update.xml", "windows-5150-expect.json"),
        (
            "5.15.2",
            "src_doc_examples",
            "windows-5152-src-doc-example-update.xml",
            "windows-5152-src-doc-example-expect.json",
        ),
        ("6.2.0", "", "windows-620-update.xml", "windows-620-expect.json"),
    ],
)
def test_list_qt_modules(monkeypatch, version: str, extension: str, in_file: str, expect_out_file: str):
    archive_id = ArchiveId("qt", "windows", "desktop")
    _xml = (Path(__file__).parent / "data" / in_file).read_text("utf-8")
    expect = json.loads((Path(__file__).parent / "data" / expect_out_file).read_text("utf-8"))

    monkeypatch.setattr(MetadataFactory, "fetch_http", lambda self, _: _xml)

    for arch in expect["architectures"]:
        modules = MetadataFactory(archive_id).fetch_modules(Version(version), arch)
        assert modules == sorted(expect["modules_by_arch"][arch])


@pytest.fixture
def win_5152_sde_xml_file() -> str:
    return (Path(__file__).parent / "data" / "windows-5152-src-doc-example-update.xml").read_text("utf-8")


def win_5152_sde_expected(cmd_type: str, query_type: str) -> Set[str]:
    assert cmd_type in ("src", "doc", "examples")
    assert query_type in ("archives", "modules")
    _json = json.loads((Path(__file__).parent / "data/windows-5152-src-doc-example-expect.json").read_text("utf-8"))
    return set(_json[cmd_type][query_type])


@pytest.mark.parametrize(
    "cmd_type, host, version, expected",
    [
        (
            _cmd_type,
            "windows",
            "5.15.2",
            win_5152_sde_expected(_cmd_type, "archives"),
        )
        for _cmd_type in ("src", "doc", "examples")
    ],
)
def test_list_src_doc_examples_archives(
    monkeypatch, win_5152_sde_xml_file, cmd_type: str, host: str, version: str, expected: Set[str]
):
    monkeypatch.setattr(MetadataFactory, "fetch_http", lambda self, _: win_5152_sde_xml_file)

    archive_id = ArchiveId("qt", host, "desktop")
    archives = set(MetadataFactory(archive_id).fetch_archives_sde(cmd_type, Version(version)))
    assert archives == expected


@pytest.mark.parametrize(
    "cmd_type, host, version, expected",
    [
        (
            _cmd_type,
            "windows",
            "5.15.2",
            win_5152_sde_expected(_cmd_type, "modules"),
        )
        for _cmd_type in ("doc", "examples")
    ],
)
def test_list_src_doc_examples_modules(
    monkeypatch, win_5152_sde_xml_file, cmd_type: str, host: str, version: str, expected: Set[str]
):
    monkeypatch.setattr(MetadataFactory, "fetch_http", lambda self, _: win_5152_sde_xml_file)

    archive_id = ArchiveId("qt", host, "desktop")
    modules = set(MetadataFactory(archive_id).fetch_modules_sde(cmd_type, Version(version)))
    assert modules == expected


@pytest.mark.parametrize(
    "command, expected",
    (
        ("list-src windows 5.15.2", win_5152_sde_expected("src", "archives")),
        ("list-doc windows 5.15.2", win_5152_sde_expected("doc", "archives")),
        ("list-example windows 5.15.2", win_5152_sde_expected("examples", "archives")),
        ("list-doc windows 5.15.2 --modules", win_5152_sde_expected("doc", "modules")),
        ("list-example windows 5.15.2 --modules", win_5152_sde_expected("examples", "modules")),
    ),
)
def test_list_src_doc_examples_cli(monkeypatch, capsys, win_5152_sde_xml_file, command: str, expected: Set[str]):
    def mock_fetch(self, rest_of_url):
        assert rest_of_url == "online/qtsdkrepository/windows_x86/desktop/qt5_5152_src_doc_examples/Updates.xml"
        return win_5152_sde_xml_file

    monkeypatch.setattr(MetadataFactory, "fetch_http", mock_fetch)

    cli = Cli()
    assert 0 == cli.run(command.split())
    out, err = capsys.readouterr()
    assert not err
    out_set = set(out.strip().split())
    assert out_set == expected


@pytest.mark.parametrize(
    "version, arch, modules_to_query, modules_failed_query",
    (
        ("5.14.0", "win32_mingw73", [], []),
        ("5.14.0", "win32_mingw73", ["qtcharts"], []),
        ("5.14.0", "win32_mingw73", ["all"], []),
        ("5.14.0", "win32_mingw73", ["debug_info"], ["debug_info"]),
        ("5.14.0", "win64_msvc2017_64", [], []),
        ("5.14.0", "win64_msvc2017_64", ["debug_info"], []),
    ),
)
def test_list_archives(
    monkeypatch, capsys, version: str, arch: str, modules_to_query: List[str], modules_failed_query: List[str]
):
    archive_id = ArchiveId("qt", "windows", "desktop")
    in_file = f"{archive_id.host}-{version.replace('.', '')}-update.xml"
    expect_out_file = f"{archive_id.host}-{version.replace('.', '')}-expect.json"
    _xml = (Path(__file__).parent / "data" / in_file).read_text("utf-8")
    monkeypatch.setattr(MetadataFactory, "fetch_http", lambda self, _: _xml)
    expect = json.loads((Path(__file__).parent / "data" / expect_out_file).read_text("utf-8"))

    if not modules_to_query:
        expected_qt_archives = expect["qt_base_pkgs_by_arch"][arch]["DownloadableArchives"]
        expected = set([arc.split("-")[0] for arc in expected_qt_archives])
    else:
        expected_mod_metadata = expect["modules_metadata_by_arch"][arch]
        if "all" not in modules_to_query:
            expected_mod_metadata = [mod for mod in expected_mod_metadata if mod["Name"].split(".")[-2] in modules_to_query]
        expected = set([arc.split("-")[0] for mod in expected_mod_metadata for arc in mod["DownloadableArchives"]])

    archives_query = [version, arch, *modules_to_query]
    cli_args = ["list-qt", "windows", "desktop", "--archives", *archives_query]
    if not modules_failed_query:
        meta = set(MetadataFactory(archive_id, archives_query=archives_query).getList())
        assert meta == expected

        cli = Cli()
        assert 0 == cli.run(cli_args)
        out, err = capsys.readouterr()
        assert out.rstrip() == " ".join(sorted(expected))
        return

    expected_err_msg = f"The requested modules were not located: {sorted(modules_failed_query)}"
    with pytest.raises(CliInputError) as err:
        MetadataFactory(archive_id, archives_query=archives_query).getList()
    assert err.type == CliInputError
    assert format(err.value).startswith(expected_err_msg)

    cli = Cli()
    assert 1 == cli.run(cli_args)
    out, err = capsys.readouterr()
    assert expected_err_msg in err


def test_list_archives_insufficient_args(capsys):
    cli = Cli()
    assert 1 == cli.run("list-qt mac desktop --archives 5.14.0".split())
    out, err = capsys.readouterr()
    assert err.strip() == "ERROR   : The '--archives' flag requires a 'QT_VERSION' and an 'ARCHITECTURE' parameter."


@pytest.mark.parametrize(
    "xml_content",
    (
        "<Updates><PackageUpdate><badname></badname></PackageUpdate></Updates>",
        "<Updates><PackageUpdate><Name></Name></PackageUpdate></Updates>",
        "<Updates></PackageUpdate><PackageUpdate></Updates><Name></Name>",
    ),
)
def test_list_archives_bad_xml(monkeypatch, xml_content: str):
    archive_id = ArchiveId("qt", "windows", "desktop")
    archives_query = ["5.15.2", "win32_mingw81", "qtcharts"]

    monkeypatch.setattr(MetadataFactory, "fetch_http", lambda self, _: xml_content)
    with pytest.raises(ArchiveListError) as e:
        MetadataFactory(archive_id, archives_query=archives_query).getList()
    assert e.type == ArchiveListError


@pytest.mark.parametrize(
    "host, target, tool_name",
    [
        ("mac", "desktop", "tools_cmake"),
        ("mac", "desktop", "tools_ifw"),
        ("mac", "desktop", "tools_qtcreator"),
    ],
)
def test_tool_modules(monkeypatch, host: str, target: str, tool_name: str):
    archive_id = ArchiveId("tools", host, target)
    in_file = "{}-{}-{}-update.xml".format(host, target, tool_name)
    expect_out_file = "{}-{}-{}-expect.json".format(host, target, tool_name)
    _xml = (Path(__file__).parent / "data" / in_file).read_text("utf-8")
    expect = json.loads((Path(__file__).parent / "data" / expect_out_file).read_text("utf-8"))

    monkeypatch.setattr(MetadataFactory, "fetch_http", lambda self, _: _xml)

    modules = MetadataFactory(archive_id, tool_name=tool_name).getList()
    assert modules == expect["modules"]

    table = MetadataFactory(archive_id, tool_name=tool_name, is_long_listing=True).getList()
    assert table._rows(table.long_heading_keys) == expect["long_listing"]


@pytest.mark.parametrize(
    "host, target, version, arch",
    [
        ["windows", "desktop", "5.14.0", arch]
        for arch in ["win32_mingw73", "win32_msvc2017", "win64_mingw73", "win64_msvc2015_64", "win64_msvc2017_64"]
    ],
)
def test_long_qt_modules(monkeypatch, host: str, target: str, version: str, arch: str):
    archive_id = ArchiveId("qt", host, target)
    in_file = f"{host}-{version.replace('.', '')}-update.xml"
    expect_out_file = f"{host}-{version.replace('.', '')}-expect.json"
    _xml = (Path(__file__).parent / "data" / in_file).read_text("utf-8")
    expect = json.loads((Path(__file__).parent / "data" / expect_out_file).read_text("utf-8"))

    monkeypatch.setattr(MetadataFactory, "fetch_http", lambda self, _: _xml)

    table = MetadataFactory(archive_id, modules_query=ModulesQuery(version, arch), is_long_listing=True).getList()
    assert table._rows(table.long_heading_keys) == expect["modules_long_by_arch"][arch]


def expected_windows_desktop_plus_wasm_5140(is_wasm_threaded: bool) -> Dict:
    if is_wasm_threaded:
        input_filenames = (
            "windows-5140-expect.json",
            "windows-650-wasm-single-expect.json",
            "windows-650-wasm-multi-expect.json",
        )
    else:
        input_filenames = "windows-5140-expect.json", "windows-5140-wasm-expect.json"
    to_join = [json.loads((Path(__file__).parent / "data" / f).read_text("utf-8")) for f in input_filenames]
    return {
        "architectures": [arch for _dict in to_join for arch in _dict["architectures"]],
        "modules_by_arch": {k: v for _dict in to_join for k, v in _dict["modules_by_arch"].items()},
    }


@pytest.mark.parametrize(
    "args, is_wasm_threaded, expect",
    (
        ("--modules latest win64_msvc2017_64", False, ["modules_by_arch", "win64_msvc2017_64"]),
        ("--spec 5.14 --modules latest win64_msvc2017_64", False, ["modules_by_arch", "win64_msvc2017_64"]),
        ("--modules 5.14.0 win32_mingw73", False, ["modules_by_arch", "win32_mingw73"]),
        ("--modules 5.14.0 win32_msvc2017", False, ["modules_by_arch", "win32_msvc2017"]),
        ("--modules 5.14.0 win64_mingw73", False, ["modules_by_arch", "win64_mingw73"]),
        ("--modules 5.14.0 win64_msvc2015_64", False, ["modules_by_arch", "win64_msvc2015_64"]),
        ("--modules 5.14.0 win64_msvc2017_64", False, ["modules_by_arch", "win64_msvc2017_64"]),
        ("--modules 6.5.0 wasm_singlethread", True, ["modules_by_arch", "wasm_singlethread"]),
        ("--modules 6.5.0 wasm_multithread", True, ["modules_by_arch", "wasm_multithread"]),
        ("--arch latest", True, ["architectures"]),
        ("--spec 5.14 --arch latest", False, ["architectures"]),
        ("--arch 5.14.0", False, ["architectures"]),
    ),
)
def test_list_qt_cli(
    monkeypatch,
    capsys,
    args: str,
    is_wasm_threaded: bool,
    expect: Union[Set[str], List[str]],
):
    version_string_to_replace = "qt5.5140"
    if isinstance(expect, list):
        # In this case, 'expect' is a list of keys to follow to the expected values.
        expected_dict = expected_windows_desktop_plus_wasm_5140(is_wasm_threaded)
        for key in expect:  # Follow the chain of keys to the list of values.
            expected_dict = expected_dict[key]
        assert isinstance(expected_dict, list)
        expect_set = set(expected_dict)
    else:
        expect_set = expect
    assert isinstance(expect_set, set)

    def _mock_fetch_http(_, rest_of_url, *args, **kwargs: str) -> str:
        htmltext = (Path(__file__).parent / "data" / "windows-desktop.html").read_text("utf-8")
        if not rest_of_url.endswith("Updates.xml"):
            return htmltext

        def get_xml_filename() -> str:
            if rest_of_url.endswith("_wasm/Updates.xml"):
                return "windows-5140-wasm-update.xml"
            elif rest_of_url.endswith("_wasm_singlethread/Updates.xml"):
                return "windows-650-wasm-single-update.xml"
            elif rest_of_url.endswith("_wasm_multithread/Updates.xml"):
                return "windows-650-wasm-multi-update.xml"
            else:
                return "windows-5140-update.xml"

        xmltext = (Path(__file__).parent / "data" / get_xml_filename()).read_text("utf-8")
        # If we are serving an Updates.xml, `aqt list` will look for a Qt version number.
        # We will replace the version numbers in the file with the requested version.
        match = re.search(r"qt(\d)_(\d+)", rest_of_url)
        assert match
        major, version_nodot = match.groups()
        desired_version_string = f"qt{major}.{version_nodot}"
        return xmltext.replace(version_string_to_replace, desired_version_string)

    monkeypatch.setattr(MetadataFactory, "fetch_http", _mock_fetch_http)

    cli = Cli()
    cli.run(["list-qt", "windows", "desktop", *args.split()])
    out, err = capsys.readouterr()
    output_set = set(out.strip().split())
    assert output_set == expect_set


def test_list_missing_wasm_updates(monkeypatch, capsys):
    """Require that MetadataFactory is resilient to missing wasm updates.xml files"""
    data_dir = Path(__file__).parent / "data"
    expect = set(json.loads((data_dir / "windows-620-expect.json").read_text("utf-8"))["architectures"])

    def _mock_fetch_http(_, rest_of_url, *args, **kwargs: str) -> str:
        htmltext = (Path(__file__).parent / "data" / "windows-desktop.html").read_text("utf-8")
        if rest_of_url.endswith("windows_x86/desktop/"):
            return htmltext
        elif rest_of_url.endswith("windows_x86/desktop/qt6_620/Updates.xml"):
            return (data_dir / "windows-620-update.xml").read_text("utf-8")
        else:
            raise ArchiveDownloadError(f"No such file at {rest_of_url}")

    monkeypatch.setattr(MetadataFactory, "fetch_http", _mock_fetch_http)

    cli = Cli()
    rv = cli.run("list-qt windows desktop --arch 6.2.0".split())
    assert rv == 0
    out, err = capsys.readouterr()
    assert set(out.strip().split()) == expect


@pytest.mark.parametrize(
    "qt_ver_str, expect_set", (("6.2.0", {"android_x86", "android_x86_64", "android_armv7", "android_arm64_v8a"}),)
)
def test_list_android_arches(monkeypatch, capsys, qt_ver_str: str, expect_set: Set[str]):
    host, target = "windows", "android"

    def _mock_fetch_http(_, rest_of_url, *args, **kwargs: str) -> str:
        assert rest_of_url.endswith("Updates.xml"), f"Fetched unexpected file at {rest_of_url}"

        xmltext = (Path(__file__).parent / "data" / "windows-620-android-armv7-update.xml").read_text("utf-8")
        # If we are serving an Updates.xml, `aqt list` will look for a Qt version number.
        # We will replace the version numbers in the file with the requested version.
        match = re.search(r"qt\d_\d+_(?P<arch>\w+)/Updates\.xml$", rest_of_url)
        assert match
        return xmltext.replace("armv7", match.group("arch"))

    monkeypatch.setattr(MetadataFactory, "fetch_http", _mock_fetch_http)

    # Unit test:
    archive_id = ArchiveId("qt", host, target)
    actual_arches = MetadataFactory(archive_id).fetch_arches(Version(qt_ver_str))
    assert set(actual_arches) == expect_set

    # Integration test:
    cli = Cli()
    cli.run(["list-qt", host, target, "--arch", qt_ver_str])
    out, err = capsys.readouterr()
    output_set = set(out.strip().split())
    assert output_set == expect_set


@pytest.mark.parametrize(
    "host, has_wasm",
    (("linux", True), ("windows", False), ("mac", False)),
)
def test_list_desktop_arches_qt5130(monkeypatch, capsys, host: str, has_wasm: bool):
    """Tests the edge case of desktop Qt 5.13.0, in which "wasm_32" is a valid arch for Linux but not win/mac."""

    # Generate some mock Updates.xml files. The host mismatch is not important; we just want to check for arch==wasm
    updates_xmltext, wasm_updates_xmltext = [
        (Path(__file__).parent / "data" / filename).read_text("utf-8").replace("qt5.5140", "qt5.5130")
        for filename in ("windows-5140-update.xml", "windows-5140-wasm-update.xml")
    ]

    def _mock_fetch_http(_, rest_of_url, *args, **kwargs: str) -> str:
        if rest_of_url == "online/qtsdkrepository/linux_x64/desktop/qt5_5130_wasm/Updates.xml":
            return wasm_updates_xmltext
        elif rest_of_url.endswith("/desktop/qt5_5130/Updates.xml"):
            return updates_xmltext
        else:
            assert False, f"Fetched an unexpected file at '{rest_of_url}'"

    monkeypatch.setattr(MetadataFactory, "fetch_http", _mock_fetch_http)

    cli = Cli()
    cli.run(["list-qt", host, "desktop", "--arch", "5.13.0"])
    out, err = capsys.readouterr()
    output_set = set(out.strip().split())
    if has_wasm:
        assert "wasm_32" in output_set
    else:
        assert "wasm_32" not in output_set


@pytest.mark.parametrize(
    "cmd, host, expect",
    (
        ("list-qt", "windows", {"desktop", "android", "winrt"}),
        ("list-qt", "linux", {"desktop", "android"}),
        ("list-qt", "mac", {"desktop", "android", "ios"}),
        ("list-tool", "windows", {"desktop", "android", "winrt"}),
        ("list-tool", "linux", {"desktop", "android"}),
        ("list-tool", "mac", {"desktop", "android", "ios"}),
    ),
)
def test_list_targets(capsys, cmd: str, host: str, expect: Set[str]):
    cli = Cli()
    cli.run([cmd, host])
    out, err = capsys.readouterr()
    output_set = set(out.strip().split())
    assert output_set == expect


@pytest.mark.parametrize(
    "cmd, host, target",
    (
        ("list-qt", "windows", "ios"),
        ("list-qt", "linux", "ios"),
        ("list-qt", "linux", "winrt"),
        ("list-qt", "mac", "winrt"),
        ("list-tool", "windows", "ios"),
        ("list-tool", "linux", "ios"),
        ("list-tool", "linux", "winrt"),
        ("list-tool", "mac", "winrt"),
    ),
)
def test_list_wrong_target(capsys, cmd: str, host: str, target: str):
    expect = f"ERROR   : '{target}' is not a valid target for host '{host}'"

    cli = Cli()
    return_code = cli.run([cmd, host, target])
    out, err = capsys.readouterr()
    assert return_code == 1
    assert err.strip() == expect


@pytest.mark.parametrize(
    "cmd, spec",
    (
        ("list-qt", "not a spec"),
        ("list-qt", "1...3"),
        ("list-qt", ""),
        ("list-qt", ">3 <5"),
    ),
)
def test_invalid_spec(capsys, cmd: str, spec: str):
    expect_prefix = f"ERROR   : Invalid version specification: '{spec}'"
    host, target = "linux", "desktop"
    cli = Cli()
    return_code = cli.run([cmd, host, target, "--spec", spec])
    out, err = capsys.readouterr()
    assert return_code == 1
    assert err.strip().startswith(expect_prefix)


@pytest.mark.parametrize(
    "simple_spec, expected_name",
    (
        (SimpleSpec("*"), "mytool.999"),
        (SimpleSpec(">3.5"), "mytool.999"),
        (SimpleSpec("3.5.5"), "mytool.355"),
        (SimpleSpec("<3.5"), "mytool.300"),
        (SimpleSpec("<=3.5"), "mytool.355"),
        (SimpleSpec("<=3.5.0"), "mytool.350"),
        (SimpleSpec(">10"), None),
    ),
)
def test_list_choose_tool_by_version(simple_spec, expected_name):
    tools_data = {
        "mytool.999": {"Version": "9.9.9", "Name": "mytool.999"},
        "mytool.355": {"Version": "3.5.5", "Name": "mytool.355"},
        "mytool.350": {"Version": "3.5.0", "Name": "mytool.350"},
        "mytool.300": {"Version": "3.0.0", "Name": "mytool.300"},
    }
    item = MetadataFactory.choose_highest_version_in_spec(tools_data, simple_spec)
    if item is not None:
        assert item["Name"] == expected_name
    else:
        assert expected_name is None


mac_qt = ArchiveId("qt", "mac", "desktop")
wrong_tool_name_msg = "Please use 'aqt list-tool mac desktop' to check what tools are available."
wrong_qt_version_msg = "Please use 'aqt list-qt mac desktop' to show versions of Qt available."
wrong_arch_msg = "Please use 'aqt list-qt mac desktop --arch <QT_VERSION>' to list valid architectures."


@pytest.mark.parametrize(
    "meta, expected_message",
    (
        (MetadataFactory(mac_qt), []),
        (
            MetadataFactory(mac_qt, spec=SimpleSpec("5.0")),
            ["Please use 'aqt list-qt mac desktop' to check that versions of qt exist within the spec '5.0'."],
        ),
        (
            MetadataFactory(ArchiveId("tools", "mac", "desktop"), tool_name="ifw"),
            [wrong_tool_name_msg],
        ),
        (
            MetadataFactory(mac_qt, architectures_ver="1.2.3"),
            [wrong_qt_version_msg],
        ),
        (
            MetadataFactory(mac_qt, modules_query=ModulesQuery("1.2.3", "clang_64")),
            [wrong_qt_version_msg, wrong_arch_msg],
        ),
        (
            MetadataFactory(mac_qt, archives_query=["1.2.3", "clang_64", "a module", "another module"]),
            [
                "Please use 'aqt list-qt mac desktop' to show versions of Qt available.",
                "Please use 'aqt list-qt mac desktop --arch <QT_VERSION>' to show architectures available.",
                "Please use 'aqt list-qt mac desktop --modules <QT_VERSION>' to show modules available.",
            ],
        ),
        (
            MetadataFactory(mac_qt, archives_query=["1.2.3", "clang_64"]),
            [
                "Please use 'aqt list-qt mac desktop' to show versions of Qt available.",
                "Please use 'aqt list-qt mac desktop --arch <QT_VERSION>' to show architectures available.",
            ],
        ),
        (
            MetadataFactory(mac_qt, spec=SimpleSpec("<5.9")),
            ["Please use 'aqt list-qt mac desktop' to check that versions of qt exist within the spec '<5.9'."],
        ),
        (
            MetadataFactory(ArchiveId("tools", "mac", "desktop"), tool_name="ifw"),
            [wrong_tool_name_msg],
        ),
        (
            MetadataFactory(mac_qt, architectures_ver="1.2.3"),
            [wrong_qt_version_msg],
        ),
        (
            MetadataFactory(mac_qt, modules_query=ModulesQuery("1.2.3", "clang_64")),
            [wrong_qt_version_msg, wrong_arch_msg],
        ),
    ),
)
def test_suggested_follow_up(meta: MetadataFactory, expected_message: str):
    assert suggested_follow_up(meta) == expected_message


def test_format_suggested_follow_up():
    suggestions = [
        "Please use 'aqt list-tool mac desktop --modules <QT_VERSION>' to list valid modules.",
        "Please use 'aqt list-tool mac desktop' to check what tools are available.",
    ]
    expected = (
        "==============================Suggested follow-up:==============================\n"
        "* Please use 'aqt list-tool mac desktop --modules <QT_VERSION>' to list valid modules.\n"
        "* Please use 'aqt list-tool mac desktop' to check what tools are available."
    )
    ex = AqtException("msg", suggested_action=suggestions)
    assert ex._format_suggested_follow_up() == expected


def test_format_suggested_follow_up_empty():
    ex = AqtException("msg", suggested_action=[])
    assert format(ex) == "msg"


@pytest.mark.parametrize(
    "meta, expect",
    (
        (
            MetadataFactory(ArchiveId("qt", "mac", "desktop"), spec=SimpleSpec("5.42")),
            "qt/mac/desktop with spec 5.42",
        ),
        (
            MetadataFactory(ArchiveId("qt", "mac", "desktop"), spec=SimpleSpec("5.42")),
            "qt/mac/desktop with spec 5.42",
        ),
        (MetadataFactory(ArchiveId("qt", "mac", "desktop")), "qt/mac/desktop"),
        (
            MetadataFactory(ArchiveId("qt", "mac", "desktop")),
            "qt/mac/desktop",
        ),
    ),
)
def test_list_describe_filters(meta: MetadataFactory, expect: str):
    assert meta.describe_filters() == expect


@pytest.mark.parametrize(
    "archive_id, spec, version_str, arch, expect",
    (
        (mac_qt, None, "5.12.42", None, Version("5.12.42")),
        (
            mac_qt,
            None,
            "not a version",
            None,
            CliInputError("Invalid version string: 'not a version'"),
        ),
        (mac_qt, SimpleSpec("5"), "latest", None, Version("5.15.2")),
        (
            mac_qt,
            SimpleSpec("5.0"),
            "latest",
            None,
            CliInputError("There is no latest version of Qt with the criteria 'qt/mac/desktop with spec 5.0'"),
        ),
        (mac_qt, SimpleSpec("<6.2.0"), "latest", "wasm_32", Version("5.15.2")),
    ),
)
def test_list_to_version(monkeypatch, archive_id, spec, version_str, arch: Optional[str], expect):
    _html = (Path(__file__).parent / "data" / "mac-desktop.html").read_text("utf-8")
    monkeypatch.setattr(MetadataFactory, "fetch_http", lambda *args, **kwargs: _html)

    if isinstance(expect, Exception):
        with pytest.raises(CliInputError) as error:
            MetadataFactory(archive_id, spec=spec)._to_version(version_str, arch)
        assert error.type == CliInputError
        assert str(expect) == str(error.value)
    else:
        assert MetadataFactory(archive_id, spec=spec)._to_version(version_str, arch) == expect


def test_list_fetch_tool_by_simple_spec(monkeypatch):
    update_xml = (Path(__file__).parent / "data" / "windows-desktop-tools_vcredist-update.xml").read_text("utf-8")
    monkeypatch.setattr(MetadataFactory, "fetch_http", lambda self, _: update_xml)

    expect_json = (Path(__file__).parent / "data" / "windows-desktop-tools_vcredist-expect.json").read_text("utf-8")
    expected = json.loads(expect_json)["modules_data"]

    def check(actual, expect):
        for key in (
            "Description",
            "DisplayName",
            "DownloadableArchives",
            "ReleaseDate",
            "SHA1",
            "Version",
            "Virtual",
        ):
            assert actual[key] == expect[key]

    meta = MetadataFactory(ArchiveId("tools", "windows", "desktop"))
    check(
        meta.fetch_tool_by_simple_spec(tool_name="tools_vcredist", simple_spec=SimpleSpec("2011")),
        expected["qt.tools.vcredist"],
    )
    check(
        meta.fetch_tool_by_simple_spec(tool_name="tools_vcredist", simple_spec=SimpleSpec("2014")),
        expected["qt.tools.vcredist_msvc2013_x86"],
    )
    nonexistent = meta.fetch_tool_by_simple_spec(tool_name="tools_vcredist", simple_spec=SimpleSpec("1970"))
    assert nonexistent is None

    # Simulate a broken Updates.xml file, with invalid versions
    highest_module_info = MetadataFactory.choose_highest_version_in_spec(
        all_tools_data={"some_module": {"Version": "not_a_version"}},
        simple_spec=SimpleSpec("*"),
    )
    assert highest_module_info is None


@pytest.mark.parametrize(
    "columns, expect",
    (
        (
            120,
            (
                "Tool Variant Name        Version         Release Date          Display Name          "
                "            Description            \n"
                "====================================================================================="
                "===================================\n"
                "qt.tools.ifw.41     4.1.1-202105261132   2021-05-26     Qt Installer Framework 4.1   "
                "The Qt Installer Framework provides\n"
                "                                                                                     "
                "a set of tools and utilities to    \n"
                "                                                                                     "
                "create installers for the supported\n"
                "                                                                                     "
                "desktop Qt platforms: Linux,       \n"
                "                                                                                     "
                "Microsoft Windows, and macOS.      \n"
            ),
        ),
        (
            80,
            "Tool Variant Name        Version         Release Date\n"
            "=====================================================\n"
            "qt.tools.ifw.41     4.1.1-202105261132   2021-05-26  \n",
        ),
        (
            0,
            "Tool Variant Name        Version         Release Date          Display Name          "
            "                                                                           Descriptio"
            "n                                                                            \n"
            "====================================================================================="
            "====================================================================================="
            "=============================================================================\n"
            "qt.tools.ifw.41     4.1.1-202105261132   2021-05-26     Qt Installer Framework 4.1   "
            "The Qt Installer Framework provides a set of tools and utilities to create installers"
            " for the supported desktop Qt platforms: Linux, Microsoft Windows, and macOS.\n",
        ),
    ),
)
def test_show_list_tools_long_ifw(capsys, monkeypatch, columns, expect):
    update_xml = (Path(__file__).parent / "data" / "mac-desktop-tools_ifw-update.xml").read_text("utf-8")
    monkeypatch.setattr(MetadataFactory, "fetch_http", lambda self, _: update_xml)

    monkeypatch.setattr(shutil, "get_terminal_size", lambda fallback: os.terminal_size((columns, 24)))

    meta = MetadataFactory(
        ArchiveId("tools", "mac", "desktop"),
        tool_name="tools_ifw",
        is_long_listing=True,
    )
    show_list(meta)
    out, err = capsys.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)
    assert out == expect


LONG_MODULES_WIN_5140_120 = (
    "   Module Name                        Display Name                     Release Date   Download Size   Installed Size\n"
    "====================================================================================================================\n"
    "qtcharts            Qt Charts for MinGW 7.3.0 32-bit                   2019-12-11     772.7K          7.8M          \n"
    "qtdatavis3d         Qt Data Visualization for MinGW 7.3.0 32-bit       2019-12-11     620.2K          5.0M          \n"
    "qtlottie            Qt Lottie Animation for MinGW 7.3.0 32-bit         2019-12-11     153.5K          968.4K        \n"
    "qtnetworkauth       Qt Network Authorization for MinGW 7.3.0 32-bit    2019-12-11     98.7K           638.6K        \n"
    "qtpurchasing        Qt Purchasing for MinGW 7.3.0 32-bit               2019-12-11     50.9K           306.8K        \n"
    "qtquick3d           Qt Quick 3D for MinGW 7.3.0 32-bit                 2019-12-11     9.8M            21.2M         \n"
    "qtquicktimeline     Qt Quick Timeline for MinGW 7.3.0 32-bit           2019-12-11     35.3K           154.1K        \n"
    "qtscript            Qt Script for MinGW 7.3.0 32-bit                   2019-12-11     1.0M            5.5M          \n"
    "qtvirtualkeyboard   Qt Virtual Keyboard for MinGW 7.3.0 32-bit         2019-12-11     2.1M            6.8M          \n"
    "qtwebglplugin       Qt WebGL Streaming Plugin for MinGW 7.3.0 32-bit   2019-12-11     201.7K          1.0M          \n"
)
LONG_MODULES_WIN_5140_80 = (
    "   Module Name                        Display Name                  \n"
    "====================================================================\n"
    "qtcharts            Qt Charts for MinGW 7.3.0 32-bit                \n"
    "qtdatavis3d         Qt Data Visualization for MinGW 7.3.0 32-bit    \n"
    "qtlottie            Qt Lottie Animation for MinGW 7.3.0 32-bit      \n"
    "qtnetworkauth       Qt Network Authorization for MinGW 7.3.0 32-bit \n"
    "qtpurchasing        Qt Purchasing for MinGW 7.3.0 32-bit            \n"
    "qtquick3d           Qt Quick 3D for MinGW 7.3.0 32-bit              \n"
    "qtquicktimeline     Qt Quick Timeline for MinGW 7.3.0 32-bit        \n"
    "qtscript            Qt Script for MinGW 7.3.0 32-bit                \n"
    "qtvirtualkeyboard   Qt Virtual Keyboard for MinGW 7.3.0 32-bit      \n"
    "qtwebglplugin       Qt WebGL Streaming Plugin for MinGW 7.3.0 32-bit\n"
)


@pytest.mark.parametrize(
    "columns, use_cli, expect",
    (
        (120, False, LONG_MODULES_WIN_5140_120),
        (80, False, LONG_MODULES_WIN_5140_80),
        (120, True, LONG_MODULES_WIN_5140_120),
        (80, True, LONG_MODULES_WIN_5140_80),
    ),
)
def test_show_list_long_qt_modules(capsys, monkeypatch, columns: int, use_cli: bool, expect: str):
    update_xml = (Path(__file__).parent / "data" / "windows-5140-update.xml").read_text("utf-8")
    monkeypatch.setattr(MetadataFactory, "fetch_http", lambda self, _: update_xml)

    cli = Cli()

    # Patching get_terminal_size prevents successful instantiation of Cli, so do it afterwards:
    monkeypatch.setattr(shutil, "get_terminal_size", lambda fallback: os.terminal_size((columns, 24)))

    if use_cli:
        return_code = cli.run("list-qt windows desktop --long-modules 5.14.0 win32_mingw73".split())
        assert return_code == 0
    else:
        meta = MetadataFactory(
            ArchiveId("qt", "windows", "desktop"),
            modules_query=ModulesQuery("5.14.0", "win32_mingw73"),
            is_long_listing=True,
        )
        show_list(meta)
    out, err = capsys.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)
    assert out == expect


def test_show_list_versions(monkeypatch, capsys):
    _html = (Path(__file__).parent / "data" / "mac-desktop.html").read_text("utf-8")
    monkeypatch.setattr(MetadataFactory, "fetch_http", lambda *args: _html)

    expect_file = Path(__file__).parent / "data" / "mac-desktop-expect.json"
    expected = "\n".join(json.loads(expect_file.read_text("utf-8"))["qt"]["qt"]) + "\n"

    show_list(MetadataFactory(mac_qt))
    out, err = capsys.readouterr()
    assert out == expected


def test_show_list_tools(monkeypatch, capsys):
    page = (Path(__file__).parent / "data" / "mac-desktop.html").read_text("utf-8")
    monkeypatch.setattr(MetadataFactory, "fetch_http", lambda *args, **kwargs: page)

    expect_file = Path(__file__).parent / "data" / "mac-desktop-expect.json"
    expect = "\n".join(json.loads(expect_file.read_text("utf-8"))["tools"]) + "\n"

    meta = MetadataFactory(ArchiveId("tools", "mac", "desktop"))
    show_list(meta)
    out, err = capsys.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)
    assert out == expect


def test_show_list_empty(monkeypatch, capsys):
    monkeypatch.setattr(MetadataFactory, "getList", lambda self: [])
    meta = MetadataFactory(ArchiveId("tools", "mac", "desktop"))
    with pytest.raises(EmptyMetadata) as error:
        show_list(meta)
    assert error.type == EmptyMetadata
    assert format(error.value) == "No tools available for this request."


@pytest.mark.parametrize(
    "exception_class, error_msg, source",
    (
        (ArchiveConnectionError, "Failure to connect to some url", "aqt.metadata.getUrl"),
        (ArchiveDownloadError, "Failure to download some xml file", "aqt.metadata.getUrl"),
    ),
)
def test_show_list_bad_connection(monkeypatch, capsys, exception_class, error_msg, source):
    def mock(*args, **kwargs):
        raise exception_class(error_msg)

    monkeypatch.setattr(source, mock)
    meta = MetadataFactory(mac_qt, spec=SimpleSpec("<5.9"))
    with pytest.raises(exception_class) as error:
        show_list(meta)
    assert error.type == exception_class
    assert format(error.value) == (
        f"{error_msg}\n"
        "==============================Suggested follow-up:==============================\n"
        "* Please use 'aqt list-qt mac desktop' to check that versions of qt exist within the spec '<5.9'."
    )


@pytest.mark.parametrize(
    "exception_class, error_msg, source", ((ArchiveConnectionError, "Failure to connect", "aqt.helper.getUrl"),)
)
def test_show_list_bad_connection_for_checksum(monkeypatch, capsys, exception_class, error_msg, source):
    def mock(*args, **kwargs):
        raise exception_class(error_msg)

    monkeypatch.setattr(source, mock)

    cli = Cli()
    cli.run("list-qt linux desktop --arch 6.4.0".split())
    out, err = capsys.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)

    assert (
        err == "ERROR   : Failed to download checksum for the file 'Updates.xml' from mirrors "
        "'['https://download.qt.io']\n"
        "==============================Suggested follow-up:==============================\n"
        "* Check your internet connection\n"
        "* Consider modifying `requests.max_retries_to_retrieve_hash` in settings.ini\n"
        "* Consider modifying `mirrors.trusted_mirrors` in settings.ini "
        "(see https://aqtinstall.readthedocs.io/en/stable/configuration.html#configuration)\n"
        "* Please use 'aqt list-qt linux desktop' to show versions of Qt available.\n"
    )


def fetch_expected_tooldata(json_filename: str) -> ToolData:
    text = (Path(__file__).parent / "data" / json_filename).read_text("utf-8")
    raw_tooldata: List[List[str]] = json.loads(text)["long_listing"]
    keys = ("Version", "ReleaseDate", "DisplayName", "Description")

    tools: Dict[str, Dict[str, str]] = {}
    for variant_name, *values in raw_tooldata:
        assert len(keys) == len(values)
        tools[variant_name] = {k: v for k, v in zip(keys, values)}

    return ToolData(tools)


@pytest.mark.parametrize(
    "host, target, tool_name",
    (
        ("mac", "desktop", "tools_cmake"),
        ("mac", "desktop", "sdktool"),
    ),
)
def test_list_tool_cli(monkeypatch, capsys, host: str, target: str, tool_name: str):
    html_file = f"{host}-{target}.html"
    xml_file = f"{host}-{target}-{tool_name}-update.xml"
    html_expect = f"{host}-{target}-expect.json"
    xml_expect = f"{host}-{target}-{tool_name}-expect.json"
    htmltext, xmltext, htmljson, xmljson = [
        (Path(__file__).parent / "data" / filename).read_text("utf-8")
        for filename in (html_file, xml_file, html_expect, xml_expect)
    ]
    expected_tools = set(json.loads(htmljson)["tools"])
    xml_data = json.loads(xmljson)
    expected_tool_modules = set(xml_data["modules"])

    def _mock_fetch_http(_, rest_of_url, *args, **kwargs: str) -> str:
        if not rest_of_url.endswith("Updates.xml"):
            return htmltext
        folder = urlparse(rest_of_url).path.split("/")[-2]
        assert folder.startswith("tools_") or folder in ["sdktool"]
        return xmltext

    monkeypatch.setattr(MetadataFactory, "fetch_http", _mock_fetch_http)

    cli = Cli()
    cli.run(["list-tool", host, target])
    out, err = capsys.readouterr()
    assert not err
    output_set = set(out.strip().split())
    assert output_set == expected_tools

    cli.run(["list-tool", host, target, tool_name])
    out, err = capsys.readouterr()
    assert not err
    output_set = set(out.strip().split())
    assert output_set == expected_tool_modules

    # Test abbreviated tool name: "aqt list-tool mac desktop ifw"
    assert tool_name.startswith("tools_") or tool_name in ["sdktool"]
    short_tool_name = tool_name[6:] if tool_name.startswith("tools_") else tool_name
    cli.run(["list-tool", host, target, short_tool_name])
    out, err = capsys.readouterr()
    assert not err
    output_set = set(out.strip().split())
    assert output_set == expected_tool_modules

    cli.run(["list-tool", host, target, tool_name, "-l"])
    out, err = capsys.readouterr()
    assert not err

    expected_tooldata = format(fetch_expected_tooldata(xml_expect))
    assert out.strip() == expected_tooldata.strip()


def test_fetch_http_ok(monkeypatch):
    html_content = b"some_html_content"
    base_url = "https://alt.baseurl.com"

    def mock_getUrl(url: str, *args, **kwargs) -> str:
        assert url.startswith(base_url)
        return str(html_content)

    monkeypatch.setattr("aqt.metadata.get_hash", lambda *args, **kwargs: hashlib.sha256(html_content).hexdigest())
    monkeypatch.setattr("aqt.metadata.getUrl", mock_getUrl)
    assert MetadataFactory(mac_qt, base_url=base_url).fetch_http("some_url") == str(html_content)


def test_fetch_http_failover(monkeypatch):
    urls_requested = set()

    def _mock(url, **kwargs):
        urls_requested.add(url)
        if len(urls_requested) <= 1:
            raise ArchiveDownloadError()
        return "some_html_content"

    monkeypatch.setattr("aqt.metadata.get_hash", lambda *args, **kwargs: hashlib.sha256(b"some_html_content").hexdigest())
    monkeypatch.setattr("aqt.metadata.getUrl", _mock)

    # Require that the first attempt failed, but the second did not
    assert MetadataFactory(mac_qt).fetch_http("some_url") == "some_html_content"
    assert len(urls_requested) == 2


@pytest.mark.parametrize("exception_on_error", (ArchiveDownloadError, ArchiveConnectionError))
def test_fetch_http_download_error(monkeypatch, exception_on_error):
    urls_requested = set()

    def _mock(url, **kwargs):
        urls_requested.add(url)
        raise exception_on_error()

    monkeypatch.setattr("aqt.metadata.get_hash", lambda *args, **kwargs: hashlib.sha256(b"some_html_content").hexdigest())
    monkeypatch.setattr("aqt.metadata.getUrl", _mock)
    with pytest.raises(exception_on_error) as e:
        MetadataFactory(mac_qt).fetch_http("some_url")
    assert e.type == exception_on_error

    # Require that a fallback url was tried
    assert len(urls_requested) == 2


@pytest.mark.parametrize(
    "host, expected, all_arches",
    (
        ("windows", "win32_mingw", ["win64_msvc2013_64", "win32_mingw", "win16_mingw"]),
        ("windows", "win32_mingw", ["win64_msvc2013_64", "win32_mingw"]),
        ("windows", "win1_mingw0", ["win64_msvc2013_64", "win_mingw900", "win1_mingw0"]),
        ("windows", "win32_mingw1", ["win64_msvc2013_64", "win32_mingw", "win32_mingw1"]),
        ("windows", "win64_mingw", ["win64_msvc2013_64", "win32_mingw900", "win64_mingw"]),
        ("windows", "win64_mingw81", ["win64_msvc2013_64", "win64_mingw81", "win64_mingw"]),
        ("windows", "win64_mingw73", ["win64_msvc2013_64", "win64_mingw53", "win64_mingw73"]),
        ("windows", "win64_mingw", ["win64_msvc2013_64", "win64_mingw", "win64_mingw"]),
        ("windows", EmptyMetadata(), []),
        ("linux", "gcc_64", ["should not fetch arches"]),
        ("mac", "clang_64", ["should not fetch arches"]),
    ),
)
def test_select_default_mingw(monkeypatch, host: str, expected: Union[str, Exception], all_arches: List[str]):
    monkeypatch.setattr("aqt.metadata.MetadataFactory.fetch_arches", lambda *args, **kwargs: all_arches)

    if isinstance(expected, Exception):
        with pytest.raises(type(expected)) as e:
            MetadataFactory(ArchiveId("qt", host, "desktop")).fetch_default_desktop_arch(Version("1.2.3"))
        assert e.type == type(expected)
    else:
        actual_arch = MetadataFactory(ArchiveId("qt", host, "desktop")).fetch_default_desktop_arch(Version("1.2.3"))
        assert actual_arch == expected


@pytest.mark.parametrize(
    "expected_result, installed_files",
    (
        ("mingw73_32", ["mingw73_32/bin/qmake.exe", "msvc2017/bin/qmake.exe"]),
        (None, ["msvc2017/bin/qmake.exe"]),
        (None, ["mingw73_win/bin/qmake.exe"]),  # Bad directory: mingw73_win does not fit the mingw naming convention
        (None, ["mingw73_32/bin/qmake", "msvc2017/bin/qmake.exe"]),
        ("mingw81_32", ["mingw73_32/bin/qmake.exe", "mingw81_32/bin/qmake.exe"]),
        ("mingw73_64", ["mingw73_64/bin/qmake.exe", "mingw73_32/bin/qmake.exe"]),
    ),
)
def test_find_installed_qt_mingw_dir(expected_result: str, installed_files: List[str]):
    qt_ver = "6.3.0"
    host = "windows"

    # Setup a mock install directory that includes some installed files
    with TemporaryDirectory() as base_dir:
        base_path = Path(base_dir)
        for file in installed_files:
            path = base_path / qt_ver / file
            path.parent.mkdir(parents=True, exist_ok=True)
            path.write_text("Mock installed file")

        actual_result = QtRepoProperty.find_installed_desktop_qt_dir(host, base_path, Version(qt_ver))
        assert (actual_result.name if actual_result else None) == expected_result



import re
import sys
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, List, Optional

import pytest

from aqt.exceptions import CliInputError
from aqt.helper import Settings
from aqt.installer import Cli
from aqt.metadata import MetadataFactory, SimpleSpec, Version


def expected_help(actual, prefix=None):
    expected = (
        "usage: aqt [-h] [-c CONFIG]\n"
        "           {install-qt,install-tool,install-doc,install-example,install-src,"
        "list-qt,list-tool,list-doc,list-example,list-src,"
        "install,tool,doc,examples,src,help,version}\n"
        "           ...\n"
        "\n"
        "Another unofficial Qt Installer.\n"
        "aqt helps you install Qt SDK, tools, examples and others\n"
        "\n"
        "option",
        "  -h, --help            show this help message and exit\n"
        "  -c CONFIG, --config CONFIG\n"
        "                        Configuration ini file.\n"
        "\n"
        "subcommands:\n"
        "  aqt accepts several subcommands:\n"
        "  install-* subcommands are commands that install components\n"
        "  list-* subcommands are commands that show available components\n"
        "  \n"
        "  commands {install|tool|src|examples|doc} are deprecated and marked for "
        "removal\n"
        "\n"
        "  {install-qt,install-tool,install-doc,install-example,install-src,list-qt,"
        "list-tool,list-doc,list-example,list-src,"
        "install,tool,doc,examples,src,help,version}\n"
        "                        Please refer to each help message by using '--help' "
        "with each subcommand\n",
    )
    if prefix is not None:
        return actual.startswith(prefix + expected[0]) and actual.endswith(expected[1])
    return actual.startswith(expected[0]) and actual.endswith(expected[1])


def test_cli_help(capsys):
    cli = Cli()
    cli.run(["help"])
    out, err = capsys.readouterr()
    assert expected_help(out)


@pytest.mark.parametrize(
    "qt_version, modules, unexpected_modules",
    (
        ("5.11.3", ["qtcharts", "qtwebengine"], []),
        ("5.11.3", ["not_exist"], ["not_exist"]),
        ("5.11.3", ["qtcharts", "qtwebengine", "not_exist"], ["not_exist"]),
        ("5.11.3", None, []),
        ("5.15.0", ["Unknown"], ["Unknown"]),
    ),
)
def test_cli_select_unexpected_modules(qt_version: str, modules: Optional[List[str]], unexpected_modules: List[str]):
    cli = Cli()
    cli._setup_settings()
    assert cli._select_unexpected_modules(qt_version, modules) == unexpected_modules

    nonexistent_qt = "5.16.0"
    assert cli._select_unexpected_modules(nonexistent_qt, modules) == sorted(modules or [])


def test_cli_check_combination():
    cli = Cli()
    cli._setup_settings()
    assert cli._check_qt_arg_combination("5.11.3", "linux", "desktop", "gcc_64")
    assert cli._check_qt_arg_combination("5.11.3", "mac", "desktop", "clang_64")
    assert not cli._check_qt_arg_combination("5.14.0", "android", "desktop", "clang_64")


def test_cli_check_version():
    cli = Cli()
    cli._setup_settings()
    assert cli._check_qt_arg_versions("5.12.0")
    assert not cli._check_qt_arg_versions("5.12")


@pytest.mark.parametrize(
    "host, target, arch, version_or_spec, expected_version, is_bad_spec",
    (
        ("windows", "desktop", "wasm_32", "6.1", None, False),
        ("windows", "desktop", "wasm_32", "5.12", None, False),
        ("windows", "desktop", "wasm_32", "5.13", Version("5.13.2"), False),
        ("windows", "desktop", "wasm_32", "5", Version("5.15.2"), False),
        ("windows", "desktop", "wasm_32", "<5.14.5", Version("5.14.2"), False),
        ("windows", "desktop", "mingw32", "6.0", Version("6.0.3"), False),
        ("windows", "winrt", "mingw32", "6", None, False),
        ("windows", "winrt", "mingw32", "bad spec", None, True),
        ("windows", "android", "android_x86", "6", Version("6.1.0"), False),
        ("windows", "desktop", "android_x86", "6", Version("6.5.0"), False),  # does not validate arch
        ("windows", "desktop", "android_fake", "6", Version("6.5.0"), False),  # does not validate arch
    ),
)
def test_cli_determine_qt_version(
    monkeypatch, host, target, arch, version_or_spec: str, expected_version: Version, is_bad_spec: bool
):
    _html = (Path(__file__).parent / "data" / f"{host}-{target}.html").read_text("utf-8")
    monkeypatch.setattr(MetadataFactory, "fetch_http", lambda *args, **kwargs: _html)
    cli = Cli()
    cli._setup_settings()

    if is_bad_spec:
        with pytest.raises(CliInputError) as e:
            Cli._determine_qt_version(version_or_spec, host, target, arch)
        assert e.type == CliInputError
        assert format(e.value) == f"Invalid version or SimpleSpec: '{version_or_spec}'\n" + SimpleSpec.usage()
    elif not expected_version:
        with pytest.raises(CliInputError) as e:
            Cli._determine_qt_version(version_or_spec, host, target, arch)
        assert e.type == CliInputError
        expect_msg = f"No versions of Qt exist for spec={version_or_spec} with host={host}, target={target}, arch={arch}"
        actual_msg = format(e.value)
        assert actual_msg == expect_msg
    else:
        ver = Cli._determine_qt_version(version_or_spec, host, target, arch)
        assert ver == expected_version


@pytest.mark.parametrize(
    "invalid_version",
    ("5.15", "five-dot-fifteen", "5", "5.5.5.5"),
)
def test_cli_invalid_version(capsys, invalid_version):
    """Checks that invalid version strings are handled properly"""

    # Ensure that invalid_version cannot be a semantic_version.Version
    with pytest.raises(ValueError):
        Version(invalid_version)

    cli = Cli()
    cli._setup_settings()

    matcher = re.compile(
        r"^INFO    : aqtinstall\(aqt\) v.* on Python 3.*\n"
        r"(.*\n)*"
        r"ERROR   :.*Invalid version: '" + invalid_version + r"'! Please use the form '5\.X\.Y'\.\n.*"
    )

    for cmd in (
        ("install", invalid_version, "mac", "desktop"),
        ("doc", invalid_version, "mac", "desktop"),
        ("list-qt", "mac", "desktop", "--arch", invalid_version),
    ):
        cli = Cli()
        assert cli.run(cmd) == 1
        out, err = capsys.readouterr()
        sys.stdout.write(out)
        sys.stderr.write(err)
        assert matcher.match(err)


@pytest.mark.parametrize(
    "version, allow_latest, allow_empty, allow_minus, expect_ok",
    (
        ("1.2.3", False, False, False, True),
        ("1.2.", False, False, False, False),
        ("latest", True, False, False, True),
        ("latest", False, False, False, False),
        ("", False, True, False, True),
        ("", False, False, False, False),
        ("1.2.3-0-123", False, False, True, True),
        ("1.2.3-0-123", False, False, False, False),
    ),
)
def test_cli_validate_version(version: str, allow_latest: bool, allow_empty: bool, allow_minus: bool, expect_ok: bool):
    if expect_ok:
        Cli._validate_version_str(version, allow_latest=allow_latest, allow_empty=allow_empty, allow_minus=allow_minus)
    else:
        with pytest.raises(CliInputError) as err:
            Cli._validate_version_str(version, allow_latest=allow_latest, allow_empty=allow_empty, allow_minus=allow_minus)
        assert err.type == CliInputError


def test_cli_check_mirror():
    cli = Cli()
    cli._setup_settings()
    assert cli._check_mirror(None)
    arg = ["install-qt", "linux", "desktop", "5.11.3", "-b", "https://download.qt.io/"]
    args = cli.parser.parse_args(arg)
    assert args.base == "https://download.qt.io/"
    assert cli._check_mirror(args.base)


@pytest.mark.parametrize(
    "arch, host, target, version, expect",
    (
        ("impossible_arch", "windows", "desktop", "6.2.0", "impossible_arch"),
        ("", "windows", "desktop", "6.2.0", None),
        (None, "windows", "desktop", "6.2.0", None),
        (None, "linux", "desktop", "6.2.0", "gcc_64"),
        (None, "mac", "desktop", "6.2.0", "clang_64"),
        (None, "mac", "ios", "6.2.0", "ios"),
        (None, "mac", "android", "6.2.0", "android"),
        (None, "mac", "android", "5.12.0", None),
        # SimpleSpec instead of Version
        ("impossible_arch", "windows", "desktop", "6.2", "impossible_arch"),
        ("", "windows", "desktop", "6.2", None),
        (None, "windows", "desktop", "6.2", None),
        (None, "linux", "desktop", "6.2", "gcc_64"),
        (None, "mac", "desktop", "6.2", "clang_64"),
        (None, "mac", "ios", "6.2", "ios"),
        (None, "mac", "android", "6.2", None),  # No way to determine arch for android target w/o version
    ),
)
def test_set_arch(arch: Optional[str], host: str, target: str, version: str, expect: Optional[str]):
    if not expect:
        with pytest.raises(CliInputError) as e:
            Cli._set_arch(arch, host, target, version)
        assert e.type == CliInputError
        assert format(e.value) == "Please supply a target architecture."
        assert e.value.should_show_help is True
    else:
        assert Cli._set_arch(arch, host, target, version) == expect


@pytest.mark.parametrize(
    "cmd, expect_msg, should_show_help",
    (
        (
            "install-qt mac ios 6.2.0 --base not_a_url",
            "The `--base` option requires a url where the path `online/qtsdkrepository` exists.",
            True,
        ),
        (
            "install-qt mac ios 6.2.0 --noarchives",
            "When `--noarchives` is set, the `--modules` option is mandatory.",
            False,
        ),
        (
            "install-qt mac ios 6.2.0 --noarchives --archives",
            "When `--noarchives` is set, the `--modules` option is mandatory.",
            False,
        ),
        (
            "install-qt mac ios 6.2.0 --noarchives --archives --modules qtcharts",
            "Options `--archives` and `--noarchives` are mutually exclusive.",
            False,
        ),
        (
            "install-src mac ios 6.2.0 --kde",
            "KDE patch: unsupported version!!",
            False,
        ),
    ),
)
def test_cli_input_errors(capsys, cmd, expect_msg, should_show_help):
    cli = Cli()
    cli._setup_settings()
    assert 1 == cli.run(cmd.split())
    out, err = capsys.readouterr()
    if should_show_help:
        assert expected_help(out)
    else:
        assert out == ""
    assert err.rstrip().endswith(expect_msg)


# These commands use the new syntax with the legacy commands
@pytest.mark.parametrize(
    "cmd",
    (
        "install linux desktop 5.10.0",
        "install linux desktop 5.10.0 gcc_64",
        "src linux desktop 5.10.0",
        "doc linux desktop 5.10.0",
        "example linux desktop 5.10.0",
        "tool windows desktop tools_ifw",
    ),
)
def test_cli_legacy_commands_with_wrong_syntax(cmd):
    cli = Cli()
    cli._setup_settings()
    with pytest.raises(SystemExit) as e:
        cli.run(cmd.split())
    assert e.type == SystemExit


@pytest.mark.parametrize(
    "cmd",
    (
        "tool windows desktop tools_ifw qt.tools.ifw.31",  # New syntax
        "tool windows desktop tools_ifw 1.2.3",
    ),
)
def test_cli_legacy_tool_new_syntax(monkeypatch, capsys, cmd):
    # These incorrect commands cannot be filtered out directly by argparse because
    # they have the correct number of arguments.
    command = cmd.split()

    expected = (
        "WARNING : The command 'tool' is deprecated and marked for removal in a future version of aqt.\n"
        "In the future, please use the command 'install-tool' instead.\n"
        "ERROR   : Invalid version: 'tools_ifw'! Please use the form '5.X.Y'.\n"
    )

    cli = Cli()
    cli._setup_settings()
    assert 1 == cli.run(command)
    out, err = capsys.readouterr()
    actual = err[err.index("\n") + 1 :]
    assert actual == expected


@pytest.mark.parametrize(
    "cmd, expect_err",
    (
        (
            "list-qt mac --extension wasm",
            "WARNING : The parameter 'extension' with value 'wasm' is deprecated "
            "and marked for removal in a future version of aqt.\n"
            "In the future, please omit this parameter.\n"
            "WARNING : The '--extension' flag will be ignored.\n",
        ),
        (
            "list-qt mac desktop --extensions 6.2.0",
            "WARNING : The parameter 'extensions' with value '6.2.0' is deprecated "
            "and marked for removal in a future version of aqt.\n"
            "In the future, please omit this parameter.\n"
            "WARNING : The '--extensions' flag will always return an empty list, "
            "because there are no useful arguments for the '--extension' flag.\n",
        ),
    ),
)
def test_cli_list_qt_deprecated_flags(capsys, cmd: str, expect_err: str):
    cli = Cli()
    cli._setup_settings()
    assert 0 == cli.run(cmd.split())
    out, err = capsys.readouterr()
    assert err == expect_err


# These commands come directly from examples in the legacy documentation
@pytest.mark.parametrize(
    "cmd",
    (
        "install 5.10.0 linux desktop",  # default arch
        "install 5.10.2 linux android android_armv7",
        "src 5.15.2 windows desktop --archives qtbase --kde",
        "doc 5.15.2 windows desktop -m qtcharts qtnetworkauth",
        "examples 5.15.2 windows desktop -m qtcharts qtnetworkauth",
        "tool linux tools_ifw 4.0 qt.tools.ifw.40",
    ),
)
def test_cli_legacy_commands_with_correct_syntax(monkeypatch, cmd):
    # Pretend to install correctly when any command is run
    for func in ("run_install_qt", "run_install_src", "run_install_doc", "run_install_example", "run_install_tool"):
        monkeypatch.setattr(Cli, func, lambda *args, **kwargs: 0)

    cli = Cli()
    cli._setup_settings()
    assert 0 == cli.run(cmd.split())


def test_cli_unexpected_error(monkeypatch, capsys):
    def _mocked_run(*args):
        raise RuntimeError("Some unexpected error")

    monkeypatch.setattr("aqt.installer.Cli.run_install_qt", _mocked_run)

    cli = Cli()
    cli._setup_settings()
    assert Cli.UNHANDLED_EXCEPTION_CODE == cli.run(["install-qt", "mac", "ios", "6.2.0"])
    out, err = capsys.readouterr()
    assert err.startswith("ERROR   : Some unexpected error")
    assert err.rstrip().endswith(
        "===========================PLEASE FILE A BUG REPORT===========================\n"
        "You have discovered a bug in aqt.\n"
        "Please file a bug report at https://github.com/miurahr/aqtinstall/issues\n"
        "Please remember to include a copy of this program's output in your report."
    )


@pytest.mark.parametrize("external_tool_exists", (True, False))
def test_set_7zip_checks_external_tool_when_specified(monkeypatch, capsys, external_tool_exists: bool):
    cli = Cli()
    cli._setup_settings()
    external = "my_7z_extractor"

    def mock_subprocess_run(args, **kwargs):
        assert args[0] == external
        if not external_tool_exists:
            raise FileNotFoundError()

    monkeypatch.setattr("aqt.installer.subprocess.run", mock_subprocess_run)
    monkeypatch.setattr("aqt.installer.EXT7Z", False)
    if external_tool_exists:
        assert external == cli._set_sevenzip(external)
    else:
        with pytest.raises(CliInputError) as err:
            cli._set_sevenzip(external)
        assert format(err.value) == format(f"Specified 7zip command executable does not exist: '{external}'")
    assert capsys.readouterr()[1] == ""


@pytest.mark.parametrize("fallback_exists", (True, False))
def test_set_7zip_uses_fallback_when_py7zr_missing(monkeypatch, capsys, fallback_exists: bool):
    cli = Cli()
    cli._setup_settings()
    external, fallback = None, Settings.zipcmd

    def mock_subprocess_run(args, **kwargs):
        assert args[0] == fallback
        if not fallback_exists:
            raise FileNotFoundError()

    monkeypatch.setattr("aqt.installer.subprocess.run", mock_subprocess_run)
    monkeypatch.setattr("aqt.installer.EXT7Z", True)
    if fallback_exists:
        assert fallback == cli._set_sevenzip(external)
    else:
        with pytest.raises(CliInputError) as err:
            cli._set_sevenzip(external)
        assert format(err.value) == format(f"Fallback 7zip command executable does not exist: '{fallback}'")
    assert f"Falling back to '{fallback}'" in capsys.readouterr()[1]


@pytest.mark.parametrize("fallback_exists", (True, False))
def test_set_7zip_chooses_p7zr_when_ext_missing(monkeypatch, capsys, fallback_exists: bool):
    cli = Cli()
    cli._setup_settings()
    external = None

    def mock_subprocess_run(args, **kwargs):
        assert False, "Should not try to run anything"

    monkeypatch.setattr("aqt.installer.subprocess.run", mock_subprocess_run)
    monkeypatch.setattr("aqt.installer.EXT7Z", False)
    assert cli._set_sevenzip(external) is None
    assert capsys.readouterr()[1] == ""


@pytest.mark.parametrize(
    "archive_dest, keep, temp_dir, expect, should_make_dir",
    (
        (None, False, "temp", "temp", False),
        (None, True, "temp", ".", False),
        ("my_archives", False, "temp", "my_archives", True),
        ("my_archives", True, "temp", "my_archives", True),
    ),
)
def test_cli_choose_archive_dest(
    monkeypatch, archive_dest: Optional[str], keep: bool, temp_dir: str, expect: str, should_make_dir: bool
):
    enclosed = {"made_dir": False}

    def mock_mkdir(*args, **kwargs):
        enclosed["made_dir"] = True

    monkeypatch.setattr("aqt.installer.Path.mkdir", mock_mkdir)

    assert Cli.choose_archive_dest(archive_dest, keep, temp_dir) == Path(expect)
    assert enclosed["made_dir"] == should_make_dir


@pytest.mark.parametrize(
    "host, target, arch, is_auto, mocked_arches, existing_arch_dirs, expect",
    (
        (  # not installed
            "windows",
            "android",
            "android_armv7",
            False,
            ["win64_mingw99"],
            ["not_mingw"],
            {"install": None, "instruct": "win64_mingw99", "use_dir": "mingw99_64"},
        ),
        (  # Alt Desktop Qt already installed
            "windows",
            "android",
            "android_armv7",
            False,
            ["win64_mingw99"],
            ["mingw128_32"],
            {"install": None, "instruct": None, "use_dir": "mingw128_32"},
        ),
        # not installed
        (
            "linux",
            "android",
            "android_armv7",
            False,
            [],
            ["gcc_32"],
            {"install": None, "instruct": "gcc_64", "use_dir": "gcc_64"},
        ),
        (  # Desktop Qt already installed
            "linux",
            "android",
            "android_armv7",
            False,
            [],
            ["gcc_64"],
            {"install": None, "instruct": None, "use_dir": "gcc_64"},
        ),
        (  # not installed
            "windows",
            "android",
            "android_armv7",
            True,
            ["win64_mingw99"],
            ["not_mingw"],
            {"install": "win64_mingw99", "instruct": None, "use_dir": "mingw99_64"},
        ),
        (  # Alt Desktop Qt already installed
            "windows",
            "android",
            "android_armv7",
            True,
            ["win64_mingw99"],
            ["mingw128_32"],
            {"install": None, "instruct": None, "use_dir": "mingw128_32"},
        ),
        # not installed
        (
            "linux",
            "android",
            "android_armv7",
            True,
            [],
            ["gcc_32"],
            {"install": "gcc_64", "instruct": None, "use_dir": "gcc_64"},
        ),
        (  # Desktop Qt already installed
            "linux",
            "android",
            "android_armv7",
            True,
            [],
            ["gcc_64"],
            {"install": None, "instruct": None, "use_dir": "gcc_64"},
        ),
        (  # MSVC arm64 with --autodesktop: should install min64_msvc2019_64
            "windows",
            "desktop",
            "win64_msvc2019_arm64",
            True,
            ["win64_mingw", "win64_msvc2019_64", "win64_msvc2019_arm64", "wasm_singlethread", "wasm_multithread"],
            ["mingw128_32"],
            {"install": "win64_msvc2019_64", "instruct": None, "use_dir": "msvc2019_64"},
        ),
        (  # MSVC arm64 without --autodesktop: should point to min64_msvc2019_64
            "windows",
            "desktop",
            "win64_msvc2019_arm64",
            False,
            ["win64_mingw", "win64_msvc2019_64", "win64_msvc2019_arm64", "wasm_singlethread", "wasm_multithread"],
            ["mingw128_32"],
            {"install": None, "instruct": "win64_msvc2019_64", "use_dir": "msvc2019_64"},
        ),
        (  # MSVC arm64 without --autodesktop, with correct target already installed
            "windows",
            "desktop",
            "win64_msvc2019_arm64",
            False,
            ["win64_mingw", "win64_msvc2019_64", "win64_msvc2019_arm64", "wasm_singlethread", "wasm_multithread"],
            ["msvc2019_64"],
            {"install": None, "instruct": None, "use_dir": "msvc2019_64"},
        ),
        (  # MSVC arm64 without --autodesktop, with wrong target already installed
            "windows",
            "desktop",
            "win64_msvc2019_arm64",
            False,
            ["win64_mingw", "win64_msvc2019_64", "win64_msvc2019_arm64", "wasm_singlethread", "wasm_multithread"],
            ["mingw128_32"],
            {"install": None, "instruct": "win64_msvc2019_64", "use_dir": "msvc2019_64"},
        ),
    ),
)
def test_get_autodesktop_dir_and_arch_non_android(
    monkeypatch,
    capsys,
    host: str,
    target: str,
    arch: str,
    is_auto: bool,
    mocked_arches: List[str],
    existing_arch_dirs: List[str],
    expect: Dict[str, str],
):
    """
    :is_auto:               Simulates passing `--autodesktop` to aqt
    :mocked_mingw:          When we ask MetadataFactory for a list of available architectures, we return this value
    :existing_arch_dirs:    Directories that contain an existing file at `arch_dir/bin/qmake`
    :expect[install]:       The archdir we expect aqt to install
    :expect[instruct]:      The architecture we expect aqt to ask the user to install
    :expect[use_dir]:       The directory that includes `bin/qmake`; we will patch files in the mobile installation
                            with this value
    """
    monkeypatch.setattr(MetadataFactory, "fetch_arches", lambda *args: mocked_arches)
    monkeypatch.setattr(Cli, "run", lambda *args: 0)

    version = "6.2.3"
    cli = Cli()
    cli._setup_settings()

    flavor = "MSVC Arm64" if arch == "win64_msvc2019_arm64" else target
    expect_msg_prefix = (
        f"You are installing the {flavor} version of Qt, "
        f"which requires that the desktop version of Qt is also installed."
    )

    with TemporaryDirectory() as temp_dir:
        base_dir = Path(temp_dir)
        for arch_dir in existing_arch_dirs:
            qmake = base_dir / version / arch_dir / f"bin/qmake{'.exe' if host == 'windows' else ''}"
            qmake.parent.mkdir(parents=True)
            qmake.write_text("exe file")
        autodesktop_arch_dir, autodesktop_arch_to_install = cli._get_autodesktop_dir_and_arch(
            is_auto, host, target, base_dir, Version(version), arch
        )
        # It should choose the correct desktop arch directory for updates
        assert autodesktop_arch_dir == expect["use_dir"]

        out, err = capsys.readouterr()
        if expect["install"]:
            assert err.strip() == f"INFO    : {expect_msg_prefix} Now installing Qt: desktop {version} {expect['install']}"
        elif expect["instruct"]:
            assert (
                err.strip() == f"WARNING : {expect_msg_prefix} You can install it with the following command:\n"
                f"          `aqt install-qt {host} desktop {version} {expect['instruct']}`"
            )
        else:
            assert err.strip() == f"INFO    : Found installed {host}-desktop Qt at {base_dir / version / expect['use_dir']}"



import os
from pathlib import Path

import pytest

from aqt.archives import QtArchives, SrcDocExamplesArchives
from aqt.helper import Settings


@pytest.fixture(autouse=True)
def setup():
    Settings.load_settings(os.path.join(os.path.dirname(__file__), "data", "settings.ini"))


@pytest.mark.parametrize(
    "os_name, version, flavor, datafile",
    (
        ("windows", "5.15.2", "doc", "windows-5152-src-doc-example-update.xml"),
        ("windows", "5.15.2", "src", "windows-5152-src-doc-example-update.xml"),
        ("windows", "5.15.2", "examples", "windows-5152-src-doc-example-update.xml"),
    ),
)
def test_parse_update_xml(monkeypatch, os_name, version, flavor, datafile):
    def _mock(self, url):
        return (Path(__file__).parent / "data" / datafile).read_text("utf-8")

    monkeypatch.setattr(QtArchives, "_download_update_xml", _mock)

    qt_archives = SrcDocExamplesArchives(flavor, os_name, "desktop", version, Settings.baseurl)
    assert qt_archives.archives is not None

    # Get packages with all extra modules
    qt_archives_all_modules = SrcDocExamplesArchives(
        flavor,
        os_name,
        "desktop",
        version,
        Settings.baseurl,
        all_extra=True,
    )
    assert qt_archives_all_modules.archives is not None

    # Extract all urls
    url_list = [item.archive_path for item in qt_archives.archives]
    url_all_modules_list = [item.archive_path for item in qt_archives_all_modules.archives]

    # Check the difference list contains only extra modules urls for target specified
    list_diff = [item for item in url_all_modules_list if item not in url_list]
    unwanted_targets = [item for item in list_diff if flavor not in item]

    # Assert if list_diff contains urls without target specified
    assert unwanted_targets == []



#!/usr/bin/env python
#
# Copyright (C) 2018 Linus Jahn <lnj@kaidan.im>
# Copyright (C) 2019-2022 Hiroshi Miura <miurahr@linux.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import posixpath
from dataclasses import dataclass, field
from logging import getLogger
from typing import Dict, Iterable, List, Optional, Set, Tuple
from xml.etree.ElementTree import Element  # noqa

from defusedxml import ElementTree

from aqt.exceptions import ArchiveDownloadError, ArchiveListError, ChecksumDownloadFailure, NoPackageFound
from aqt.helper import Settings, get_hash, getUrl, ssplit
from aqt.metadata import QtRepoProperty, Version


@dataclass
class TargetConfig:
    version: str
    target: str
    arch: str
    os_name: str


@dataclass
class QtPackage:
    name: str
    base_url: str
    archive_path: str
    archive: str
    package_desc: str
    pkg_update_name: str
    version: Optional[Version] = field(default=None)

    def __repr__(self):
        v_info = f", version={self.version}" if self.version else ""
        return f"QtPackage(name={self.name}, archive={self.archive}{v_info})"

    def __str__(self):
        v_info = f", version={self.version}" if self.version else ""
        return (
            f"QtPackage(name={self.name}, url={self.archive_path}, "
            f"archive={self.archive}, desc={self.package_desc}"
            f"{v_info})"
        )


class ModuleToPackage:
    """
    Holds a mapping of module names to a list of Updates.xml PackageUpdate names.
    For example, we could have the following:
    {"qtcharts": ["qt.qt6.620.addons.qtcharts.arch", qt.qt6.620.qtcharts.arch", qt.620.addons.qtcharts.arch",])
    It also contains a reverse mapping of PackageUpdate names to module names, so that
    lookup of a package name and removal of a module name can be done in constant time.
    Without this reverse mapping, QtArchives._parse_update_xml would run at least one
    linear search on the forward mapping for each module installed.

    The list of PackageUpdate names consists of all the possible names for the PackageUpdate.
    The naming conventions for each PackageUpdate are not predictable, so we need to maintain
    a list of possibilities. While reading Updates.xml, if we encounter any one of the package
    names on this list, we can use it to install the package "qtcharts".

    Once we have installed the package, we need to remove the package "qtcharts" from this
    mapping, so we can keep track of what still needs to be installed.
    """

    def __init__(self, initial_map: Dict[str, List[str]]):
        self._modules_to_packages: Dict[str, List[str]] = initial_map
        self._packages_to_modules: Dict[str, str] = {
            value: key for key, list_of_values in initial_map.items() for value in list_of_values
        }

    def add(self, module_name: str, package_names: List[str]):
        self._modules_to_packages[module_name] = self._modules_to_packages.get(module_name, []) + package_names
        for package_name in package_names:
            assert package_name not in self._packages_to_modules, "Detected a package name collision"
            self._packages_to_modules[package_name] = module_name

    def remove_module_for_package(self, package_name: str):
        module_name = self._packages_to_modules[package_name]
        for package_name in self._modules_to_packages[module_name]:
            self._packages_to_modules.pop(package_name)
        self._modules_to_packages.pop(module_name)

    def has_package(self, package_name: str):
        return package_name in self._packages_to_modules

    def get_modules(self) -> Iterable[str]:
        return self._modules_to_packages.keys()

    def __len__(self) -> int:
        return len(self._modules_to_packages)

    def __format__(self, format_spec) -> str:
        return str(sorted(set(self._modules_to_packages.keys())))


@dataclass
class PackageUpdate:
    """
    Data class to hold package data.
    key is its name.
    """

    name: str
    display_name: str
    description: str
    release_date: str
    full_version: str
    dependencies: Iterable[str]
    auto_dependon: Iterable[str]
    downloadable_archives: Iterable[str]
    default: bool
    virtual: bool
    base: str

    def __post_init__(self):
        for iter_of_str in self.dependencies, self.auto_dependon, self.downloadable_archives:
            assert isinstance(iter_of_str, Iterable) and not isinstance(iter_of_str, str)
        for _str in self.name, self.display_name, self.description, self.release_date, self.full_version, self.base:
            assert isinstance(_str, str)
        for boolean in self.default, self.virtual:
            assert isinstance(boolean, bool)

    @property
    def version(self):
        return Version.permissive(self.full_version)

    @property
    def arch(self):
        return self.name.split(".")[-1]

    def is_base_package(self) -> bool:
        return self.name in (
            f"qt.qt{self.version.major}.{self._version_str()}.{self.arch}",
            f"qt.{self._version_str()}.{self.arch}",
        )

    def _version_str(self) -> str:
        return ("{0.major}{0.minor}" if self.version == Version("5.9.0") else "{0.major}{0.minor}{0.patch}").format(
            self.version
        )


@dataclass(init=False)
class Updates:
    package_updates: List[PackageUpdate]

    def __init__(self):
        self.package_updates = []

    def extend(self, other):
        self.package_updates.extend(other.package_updates)

    @staticmethod
    def fromstring(base, update_xml_text: str):
        try:
            update_xml = ElementTree.fromstring(update_xml_text)
        except ElementTree.ParseError as perror:
            raise ArchiveListError(f"Downloaded metadata is corrupted. {perror}") from perror
        updates = Updates()
        for packageupdate in update_xml.iter("PackageUpdate"):
            pkg_name = updates._get_text(packageupdate.find("Name"))
            display_name = updates._get_text(packageupdate.find("DisplayName"))
            full_version = updates._get_text(packageupdate.find("Version"))
            package_desc = updates._get_text(packageupdate.find("Description"))
            release_date = updates._get_text(packageupdate.find("ReleaseDate"))
            dependencies = updates._get_list(packageupdate.find("Dependencies"))
            auto_dependon = updates._get_list(packageupdate.find("AutoDependOn"))
            archives = updates._get_list(packageupdate.find("DownloadableArchives"))
            default = updates._get_boolean(packageupdate.find("Default"))
            virtual = updates._get_boolean(packageupdate.find("Virtual"))
            updates.package_updates.append(
                PackageUpdate(
                    pkg_name,
                    display_name,
                    package_desc,
                    release_date,
                    full_version,
                    dependencies,
                    auto_dependon,
                    archives,
                    default,
                    virtual,
                    base,
                )
            )
        return updates

    def get(self, target: Optional[str] = None):
        if target is None:
            return self.package_updates
        for update in self.package_updates:
            if update.name == target:
                return update
        return None

    def get_from(self, arch: str, is_include_base: bool, target_packages: Optional[ModuleToPackage] = None):
        result = []
        for update in self.package_updates:
            # If we asked for `--noarchives`, we don't want the base module
            if not is_include_base and update.is_base_package():
                continue
            if target_packages is not None and not target_packages.has_package(update.name):
                continue
            if arch in update.name:
                result.append(update)
        return result

    def merge(self, other):
        self.package_updates.extend(other.package_updates)

    def get_depends(self, target: str) -> Iterable[str]:
        # initialize
        filo = [target]
        packages = []
        visited = []
        # dfs look-up
        while len(filo) > 0:
            next = filo.pop()
            packages.append(next)
            for entry in self.package_updates:
                if entry.name == next:
                    visited.append(next)
                    if entry.dependencies is not None:
                        for depend in entry.dependencies:
                            if depend not in visited:
                                filo.append(depend)
        return packages

    def _get_text(self, item: Optional[Element]) -> str:
        if item is not None and item.text is not None:
            return item.text
        return ""

    def _get_list(self, item: Optional[Element]) -> Iterable[str]:
        if item is not None and item.text is not None:
            return ssplit(item.text)
        else:
            return []

    def _get_boolean(self, item) -> bool:
        if "true" == item:
            return True
        else:
            return False


class QtArchives:
    """Download and hold Qt archive packages list.
    It access to download.qt.io site and get Update.xml file.
    It parse XML file and store metadata into list of QtPackage object.
    """

    def __init__(
        self,
        os_name: str,
        target: str,
        version_str: str,
        arch: str,
        base: str,
        subarchives: Optional[Iterable[str]] = None,
        modules: Optional[Iterable[str]] = None,
        all_extra: bool = False,
        is_include_base_package: bool = True,
        timeout=(5, 5),
    ):
        self.version: Version = Version(version_str)
        self.target: str = target
        self.arch: str = arch
        self.os_name: str = os_name
        self.all_extra: bool = all_extra
        self.arch_list: List[str] = [item.get("arch") for item in Settings.qt_combinations]
        self.base: str = base
        self.logger = getLogger("aqt.archives")
        self.archives: List[QtPackage] = []
        self.subarchives: Optional[Iterable[str]] = subarchives
        self.mod_list: Set[str] = set(modules or [])
        self.is_include_base_package: bool = is_include_base_package
        self.timeout = timeout
        try:
            self._get_archives()
        except ArchiveDownloadError as e:
            self.handle_missing_updates_xml(e)

    def handle_missing_updates_xml(self, e: ArchiveDownloadError):
        msg = f"Failed to locate XML data for Qt version '{self.version}'."
        help_msg = f"Please use 'aqt list-qt {self.os_name} {self.target}' to show versions available."
        raise ArchiveListError(msg, suggested_action=[help_msg]) from e

    def should_filter_archives(self, package_name: str) -> bool:
        """
        This tells us, based on the PackageUpdate.Name property, whether or not the `self.subarchives`
        list should be used to filter out archives that we are not interested in.

        If `package_name` is a base module or a debug_info module, the `subarchives` list will apply to it.
        """
        return package_name in self._base_package_names() or "debug_info" in package_name

    def _version_str(self) -> str:
        return ("{0.major}{0.minor}" if self.version == Version("5.9.0") else "{0.major}{0.minor}{0.patch}").format(
            self.version
        )

    def _arch_ext(self) -> str:
        ext = QtRepoProperty.extension_for_arch(self.arch, self.version >= Version("6.0.0"))
        return ("_" + ext) if ext else ""

    def _base_module_name(self) -> str:
        """
        This is the name for the base Qt module, whose PackageUpdate.Name property would be
        'qt.123.gcc_64' or 'qt.qt1.123.gcc_64' for Qt 1.2.3, for architecture gcc_64.
        """
        return "qt_base"

    def _base_package_names(self) -> Iterable[str]:
        """
        This is a list of all potential PackageUpdate.Name properties for the base Qt module,
        which would be 'qt.123.gcc_64' or 'qt.qt1.123.gcc_64' for Qt 1.2.3, for architecture gcc_64,
        or 'qt.123.src' or 'qt.qt1.123.src' for the source module.
        """
        return (
            f"qt.qt{self.version.major}.{self._version_str()}.{self.arch}",
            f"qt.{self._version_str()}.{self.arch}",
        )

    def _module_name_suffix(self, module: str) -> str:
        return f"{module}.{self.arch}"

    def _target_packages(self) -> ModuleToPackage:
        if self.all_extra:
            return ModuleToPackage({})
        base_package = {self._base_module_name(): list(self._base_package_names())}
        target_packages = ModuleToPackage(base_package if self.is_include_base_package else {})
        for module in self.mod_list:
            suffix = self._module_name_suffix(module)
            package_names = [
                f"qt.qt{self.version.major}.{self._version_str()}.{suffix}",
                f"qt.{self._version_str()}.{suffix}",
            ]
            if not module.startswith("addons."):
                package_names.append(f"qt.qt{self.version.major}.{self._version_str()}.addons.{suffix}")
            target_packages.add(module, package_names)
        return target_packages

    def _get_archives(self):
        self._get_archives_base(f"qt{self.version.major}_{self._version_str()}{self._arch_ext()}", self._target_packages())

    def _append_depends_tool(self, arch, tool_name):
        os_target_folder = posixpath.join(
            "online/qtsdkrepository",
            self.os_name + ("_x86" if self.os_name == "windows" else "_x64"),
            self.target,
            tool_name,
        )
        update_xml_url = posixpath.join(os_target_folder, "Updates.xml")
        update_xml_text = self._download_update_xml(update_xml_url)
        update_xml = Updates.fromstring(self.base, update_xml_text)
        self._append_tool_update(os_target_folder, update_xml, arch, None)

    def _get_archives_base(self, name, target_packages):
        os_target_folder = posixpath.join(
            "online/qtsdkrepository",
            self.os_name + ("_x86" if self.os_name == "windows" else "_x64"),
            self.target,
            # tools_ifw/
            name,
        )
        update_xml_url = posixpath.join(os_target_folder, "Updates.xml")
        update_xml_text = self._download_update_xml(update_xml_url)
        self._parse_update_xml(os_target_folder, update_xml_text, target_packages)

    def _download_update_xml(self, update_xml_path):
        """Hook for unit test."""
        if not Settings.ignore_hash:
            try:
                xml_hash = get_hash(update_xml_path, Settings.hash_algorithm, self.timeout)
            except ChecksumDownloadFailure:
                self.logger.warning(
                    "Failed to download checksum for the file 'Updates.xml'. This may happen on unofficial mirrors."
                )
                xml_hash = None
        else:
            xml_hash = None
        return getUrl(posixpath.join(self.base, update_xml_path), self.timeout, xml_hash)

    def _parse_update_xml(self, os_target_folder, update_xml_text, target_packages: Optional[ModuleToPackage]):
        if not target_packages:
            target_packages = ModuleToPackage({})
        update_xml = Updates.fromstring(self.base, update_xml_text)
        base_url = self.base
        if self.all_extra:
            package_updates = update_xml.get_from(self.arch, self.is_include_base_package)
        else:
            package_updates = update_xml.get_from(self.arch, self.is_include_base_package, target_packages)
        for packageupdate in package_updates:
            if not self.all_extra:
                target_packages.remove_module_for_package(packageupdate.name)
            should_filter_archives: bool = bool(self.subarchives) and self.should_filter_archives(packageupdate.name)

            for archive in packageupdate.downloadable_archives:
                archive_name = archive.split("-", maxsplit=1)[0]
                if should_filter_archives and self.subarchives is not None and archive_name not in self.subarchives:
                    continue
                archive_path = posixpath.join(
                    # online/qtsdkrepository/linux_x64/desktop/qt5_5150/
                    os_target_folder,
                    # qt.qt5.5150.gcc_64/
                    packageupdate.name,
                    # 5.15.0-0-202005140804qtbase-Linux-RHEL_7_6-GCC-Linux-RHEL_7_6-X86_64.7z
                    packageupdate.full_version + archive,
                )
                self.archives.append(
                    QtPackage(
                        name=archive_name,
                        base_url=base_url,
                        archive_path=archive_path,
                        archive=archive,
                        package_desc=packageupdate.description,
                        pkg_update_name=packageupdate.name,  # For testing purposes
                    )
                )
        # if we have located every requested package, then target_packages will be empty
        if not self.all_extra and len(target_packages) > 0:
            message = f"The packages {target_packages} were not found while parsing XML of package information!"
            raise NoPackageFound(message, suggested_action=self.help_msg(list(target_packages.get_modules())))

    def _append_tool_update(self, os_target_folder, update_xml, target, tool_version_str):
        packageupdate = update_xml.get(target)
        if packageupdate is None:
            message = f"The package '{self.arch}' was not found while parsing XML of package information!"
            raise NoPackageFound(message, suggested_action=self.help_msg())
        name = packageupdate.name
        named_version = packageupdate.full_version
        if tool_version_str and named_version != tool_version_str:
            message = f"The package '{self.arch}' has the version '{named_version}', not the requested '{self.version}'."
            raise NoPackageFound(message, suggested_action=self.help_msg())
        package_desc = packageupdate.description
        downloadable_archives = packageupdate.downloadable_archives
        if not downloadable_archives:
            message = f"The package '{self.arch}' contains no downloadable archives!"
            raise NoPackageFound(message)
        for archive in downloadable_archives:
            archive_path = posixpath.join(
                # online/qtsdkrepository/linux_x64/desktop/tools_ifw/
                os_target_folder,
                # qt.tools.ifw.41/
                name,
                #  4.1.1-202105261130ifw-linux-x64.7z
                f"{named_version}{archive}",
            )
            self.archives.append(
                QtPackage(
                    name=name,
                    base_url=self.base,
                    archive_path=archive_path,
                    archive=archive,
                    package_desc=package_desc,
                    pkg_update_name=name,  # Redundant
                )
            )

    def help_msg(self, missing_modules: Optional[List[str]] = None) -> List[str]:
        _missing_modules: List[str] = missing_modules or []
        base_cmd = f"aqt list-qt {self.os_name} {self.target}"
        arch = f"Please use '{base_cmd} --arch {self.version}' to show architectures available."
        mods = f"Please use '{base_cmd} --modules {self.version} <arch>' to show modules available."
        has_base_pkg: bool = self._base_module_name() in _missing_modules
        has_non_base_pkg: bool = len(list(_missing_modules)) > 1 or not has_base_pkg
        messages = []
        if has_base_pkg:
            messages.append(arch)
        if has_non_base_pkg:
            messages.append(mods)
        return messages

    def get_packages(self) -> List[QtPackage]:
        """
         It returns an archive package list.

        :return package list
        :rtype: List[QtPackage]
        """
        return self.archives

    def get_target_config(self) -> TargetConfig:
        """Get target configuration

        :return: configured target and its version with arch
        :rtype: TargetConfig object
        """
        return TargetConfig(str(self.version), self.target, self.arch, self.os_name)


class SrcDocExamplesArchives(QtArchives):
    """Hold doc/src/example archive package list."""

    def __init__(
        self,
        flavor: str,
        os_name,
        target,
        version,
        base,
        subarchives=None,
        modules=None,
        all_extra=False,
        is_include_base_package: bool = True,
        timeout=(5, 5),
    ):
        self.flavor: str = flavor
        self.target = target
        self.os_name = os_name
        self.base = base
        self.logger = getLogger("aqt.archives")
        super(SrcDocExamplesArchives, self).__init__(
            os_name,
            target,
            version,
            arch=self.flavor,
            base=base,
            subarchives=subarchives,
            modules=modules,
            all_extra=all_extra,
            is_include_base_package=is_include_base_package,
            timeout=timeout,
        )

    def _arch_ext(self) -> str:
        return "_src_doc_examples"

    def _base_module_name(self) -> str:
        """
        This is the name for the base Qt Src/Doc/Example module, whose PackageUpdate.Name
        property would be 'qt.123.examples' or 'qt.qt1.123.examples' for Qt 1.2.3 examples.
        """
        return self.flavor  # src | doc | examples

    def _module_name_suffix(self, module: str) -> str:
        return f"{self.flavor}.{module}"

    def get_target_config(self) -> TargetConfig:
        """Get target configuration.

        :return tuple of three parameter, "src_doc_examples", target and arch
        """
        return TargetConfig("src_doc_examples", self.target, self.arch, self.os_name)

    def help_msg(self, missing_modules: Optional[List[str]] = None) -> List[str]:
        _missing_modules: List[str] = missing_modules or []
        cmd_type = "example" if self.flavor == "examples" else self.flavor
        base_cmd = f"aqt list-{cmd_type} {self.os_name} {self.version}"
        mods = f"Please use '{base_cmd} --modules' to show modules available."
        has_non_base_pkg: bool = len(list(_missing_modules)) > 1
        messages = []
        if has_non_base_pkg:
            messages.append(mods)
        return messages


class ToolArchives(QtArchives):
    """Hold tool archive package list
    when installing mingw tool, argument would be
    ToolArchive(windows, desktop, 4.9.1-3, mingw)
    when installing ifw tool, argument would be
    ToolArchive(linux, desktop, 3.1.1, ifw)
    """

    def __init__(
        self,
        os_name: str,
        target: str,
        tool_name: str,
        base: str,
        version_str: Optional[str] = None,
        arch: str = "",
        timeout: Tuple[float, float] = (5, 5),
    ):
        self.tool_name = tool_name
        self.os_name = os_name
        self.logger = getLogger("aqt.archives")
        self.tool_version_str: Optional[str] = version_str
        super(ToolArchives, self).__init__(
            os_name=os_name,
            target=target,
            version_str="0.0.1",  # dummy value
            arch=arch,
            base=base,
            timeout=timeout,
        )

    def __str__(self):
        return f"ToolArchives(tool_name={self.tool_name}, version={self.version}, arch={self.arch})"

    def handle_missing_updates_xml(self, e: ArchiveDownloadError):
        msg = f"Failed to locate XML data for the tool '{self.tool_name}'."
        help_msg = f"Please use 'aqt list-tool {self.os_name} {self.target}' to show tools available."
        raise ArchiveListError(msg, suggested_action=[help_msg]) from e

    def _get_archives(self):
        self._get_archives_base(self.tool_name, None)

    def _parse_update_xml(self, os_target_folder, update_xml_text, *ignored):
        update_xml = Updates.fromstring(self.base, update_xml_text)
        self._append_tool_update(os_target_folder, update_xml, self.arch, self.tool_version_str)

    def help_msg(self, *args) -> List[str]:
        return [f"Please use 'aqt list-tool {self.os_name} {self.target} {self.tool_name}' to show tool variants available."]

    def get_target_config(self) -> TargetConfig:
        """Get target configuration.

        :return tuple of three parameter, "Tools", target and arch
        """
        return TargetConfig("Tools", self.target, self.arch, self.os_name)



"""
This sets variables for a matrix of QT versions to test downloading against with Azure Pipelines
"""
import collections
import json
import secrets as random
import re
from itertools import product
from typing import Dict, Optional

MIRRORS = [
    "https://ftp.jaist.ac.jp/pub/qtproject",
    "https://ftp1.nluug.nl/languages/qt",
    "https://mirrors.dotsrc.org/qtproject",
]


class BuildJob:
    def __init__(
        self,
        command,
        qt_version,
        host,
        target,
        arch,
        archdir,
        *,
        module=None,
        mirror=None,
        subarchives=None,
        output_dir=None,
        list_options=None,
        spec=None,
        mingw_variant: str = "",
        is_autodesktop: bool = False,
        tool_options: Optional[Dict[str, str]] = None,
        check_output_cmd: Optional[str] = None,
        emsdk_version: str = "sdk-fastcomp-1.38.27-64bit@3.1.29",
        autodesk_arch_folder: Optional[str] = None,
    ):
        self.command = command
        self.qt_version = qt_version
        self.host = host
        self.target = target
        self.arch = arch
        self.archdir = archdir
        self.module = module
        self.mirror = mirror
        self.subarchives = subarchives
        self.mingw_variant: str = mingw_variant
        self.is_autodesktop: bool = is_autodesktop
        self.list_options = list_options if list_options else {}
        self.tool_options: Dict[str, str] = tool_options if tool_options else {}
        # `steps.yml` assumes that qt_version is the highest version that satisfies spec
        self.spec = spec
        self.output_dir = output_dir
        self.check_output_cmd = check_output_cmd
        self.emsdk_version = emsdk_version
        self.autodesk_arch_folder = autodesk_arch_folder

    def qt_bindir(self, *, sep='/') -> str:
        out_dir = f"$(Build.BinariesDirectory){sep}Qt" if not self.output_dir else self.output_dir
        version_dir = "5.9" if self.qt_version == "5.9.0" else self.qt_version
        return f"{out_dir}{sep}{version_dir}{sep}{self.archdir}{sep}bin"

    def win_qt_bindir(self) -> str:
        return self.qt_bindir(sep='\\')

    def autodesk_qt_bindir(self, *, sep='/') -> str:
        out_dir = f"$(Build.BinariesDirectory){sep}Qt" if not self.output_dir else self.output_dir
        version_dir = "5.9" if self.qt_version == "5.9.0" else self.qt_version
        return f"{out_dir}{sep}{version_dir}{sep}{self.autodesk_arch_folder or self.archdir}{sep}bin"

    def win_autodesk_qt_bindir(self) -> str:
        return self.autodesk_qt_bindir(sep='\\')

    def mingw_folder(self) -> str:
        if not self.mingw_variant:
            return ""
        match = re.match(r"^win(\d+)_(mingw\d+)$", self.mingw_variant)
        return f"{match[2]}_{match[1]}"


class PlatformBuildJobs:
    def __init__(self, platform, build_jobs):
        self.platform = platform
        self.build_jobs = build_jobs


python_versions = [
    "3.9",
]

qt_versions = ["5.13.2", "5.15.2"]

linux_build_jobs = []
mac_build_jobs = []
windows_build_jobs = []

all_platform_build_jobs = [
    PlatformBuildJobs("linux", linux_build_jobs),
    PlatformBuildJobs("mac", mac_build_jobs),
    PlatformBuildJobs("windows", windows_build_jobs),
]

# Linux Desktop
for qt_version in qt_versions:
    linux_build_jobs.append(
        BuildJob("install-qt", qt_version, "linux", "desktop", "gcc_64", "gcc_64")
    )

# Mac Desktop
for qt_version in qt_versions:
    mac_build_jobs.append(
        BuildJob("install-qt", qt_version, "mac", "desktop", "clang_64", "clang_64")
    )

# Windows Desktop
windows_build_jobs.extend(
    [
        BuildJob(
            "install-qt",
            "5.15.2",
            "windows",
            "desktop",
            "win32_msvc2019",
            "msvc2019",
            mirror=random.choice(MIRRORS),
        ),
        BuildJob(
            "install-qt",
            "5.14.2",
            "windows",
            "desktop",
            "win32_mingw73",
            "mingw73_32",
            mingw_variant="win32_mingw730",
            mirror=random.choice(MIRRORS),
        ),
        BuildJob(
            "install-qt",
            "5.13.2",
            "windows",
            "desktop",
            "win64_msvc2015_64",
            "msvc2015_64",
            mirror=random.choice(MIRRORS),
        ),
        BuildJob(
            "install-qt",
            "5.15.2",
            "windows",
            "desktop",
            "win64_mingw81",
            "mingw81_64",
            mingw_variant="win64_mingw810",
            mirror=random.choice(MIRRORS),
        ),
        # Known issue with Azure-Pipelines environment: it has a pre-installed mingw81 which cause link error.
        # BuildJob('install', '5.15.0', 'windows', 'desktop', 'win32_mingw81', 'mingw81_32', mirror=MIRROR),
        BuildJob(
            "install-qt",
            "5.15.2",
            "windows",
            "desktop",
            "win64_msvc2019_64",
            "msvc2019_64",
            module="qtcharts qtnetworkauth",
            mirror=random.choice(MIRRORS),
        ),
        BuildJob(
            "install-qt",
            "5.9.0",
            "windows",
            "desktop",
            "win64_msvc2015_64",
            "msvc2015_64",
            module="qtcharts qtnetworkauth",
            mirror=random.choice(MIRRORS),
        ),
        BuildJob(
            "install-qt",
            "5.14.2",
            "windows",
            "desktop",
            "win64_mingw73",
            "mingw73_64",
            mingw_variant="win64_mingw730",
            spec=">1,<5.15",  # Don't redirect output! Must be wrapped in quotes!
            mirror=random.choice(MIRRORS),
        ),
        BuildJob(
            "install-qt",
            "6.5.2",
            "windows",
            "desktop",
            "win64_msvc2019_arm64",
            "msvc2019_arm64",
            is_autodesktop=True,    # Should install win64_msvc2019_arm64 in parallel
        ),
        BuildJob(
            # Archives stored as .zip
            "install-src", "6.4.3", "windows", "desktop", "gcc_64", "gcc_64", subarchives="qtlottie",
            # Fail the job if this path does not exist:
            check_output_cmd="ls -lh ./6.4.3/Src/qtlottie/",
        ),
    ]
)

# Extra modules test
linux_build_jobs.extend(
    [
        BuildJob(
            "install-qt",
            "5.15.2",
            "linux",
            "desktop",
            "gcc_64",
            "gcc_64",
            module="qtcharts qtnetworkauth",
        ),
        BuildJob(
            "install-qt", "5.14.2", "linux", "desktop", "gcc_64", "gcc_64", module="all"
        ),
        BuildJob(
            "install-qt",
            "5.15.2",
            "linux",
            "desktop",
            "gcc_64",
            "gcc_64",
            subarchives="qtbase qttools qt icu",
        ),
        BuildJob(
            # Archives stored as .7z
            "install-src", "6.1.0", "linux", "desktop", "gcc_64", "gcc_64", subarchives="qtlottie",
            # Fail the job if this path does not exist:
            check_output_cmd="ls -lh ./6.1.0/Src/qtlottie/",
        ),
        BuildJob(
            # Archives stored as .tar.gz
            "install-src", "6.4.3", "linux", "desktop", "gcc_64", "gcc_64", subarchives="qtlottie",
            # Fail the job if this path does not exist:
            check_output_cmd="ls -lh ./6.4.3/Src/qtlottie/",
        ),
        # Should install the `qtlottie` module, even though the archive `qtlottieanimation` is not specified:
        BuildJob(
            "install-doc", "6.1.0", "linux", "desktop", "gcc_64", "gcc_64",
            subarchives="qtdoc", module="qtlottie",
            # Fail the job if these paths do not exist:
            check_output_cmd="ls -lh ./Docs/Qt-6.1.0/qtdoc/ ./Docs/Qt-6.1.0/qtlottieanimation/",
        ),
        # Should install the `qtcharts` module, even though the archive `qtcharts` is not specified:
        BuildJob(
            "install-example", "6.1.0", "linux", "desktop", "gcc_64", "gcc_64",
            subarchives="qtdoc", module="qtcharts",
            # Fail the job if these paths do not exist:
            check_output_cmd="ls -lh ./Examples/Qt-6.1.0/charts/ ./Examples/Qt-6.1.0/demos/ ./Examples/Qt-6.1.0/tutorials/",
        ),
        # test for list commands
        BuildJob('list', '5.15.2', 'linux', 'desktop', 'gcc_64', '', spec="<6", list_options={'HAS_WASM': "True"}),
        BuildJob('list', '6.1.0', 'linux', 'desktop', 'gcc_64', '', spec=">6.0,<6.1.1", list_options={'HAS_WASM': "False"}),
        BuildJob('list', '6.1.0', 'linux', 'android', 'android_armv7', '', spec=">6.0,<6.1.1", list_options={}),
        # tests run on linux but query data about other platforms
        BuildJob('list', '5.14.1', 'mac', 'ios', 'ios', '', spec="<=5.14.1", list_options={}),
        BuildJob('list', '5.13.1', 'windows', 'winrt', 'win64_msvc2015_winrt_x64', '', spec=">5.13.0,<5.13.2", list_options={}),
    ]
)
mac_build_jobs.extend(
    [
        BuildJob(
            "install-qt",
            "6.2.0",
            "mac",
            "desktop",
            "clang_64",
            "macos",
            module="qtcharts qtnetworkauth",
        ),
        BuildJob(
            "install-qt",
            "5.14.2",
            "mac",
            "desktop",
            "clang_64",
            "clang_64",
            module="qtcharts qtnetworkauth",
        ),
    ]
)

# WASM
linux_build_jobs.append(
    BuildJob("install-qt", "5.14.2", "linux", "desktop", "wasm_32", "wasm_32")
)
linux_build_jobs.append(
    BuildJob("install-qt", "6.4.0", "linux", "desktop", "wasm_32", "wasm_32",
             is_autodesktop=True, emsdk_version="sdk-3.1.14-64bit", autodesk_arch_folder="gcc_64")
)
for job_queue, host, desk_arch in (
    (linux_build_jobs, "linux", "gcc_64"),
    (mac_build_jobs, "mac", "clang_64"),
    (windows_build_jobs, "windows", "mingw_64"),
):
    for wasm_arch in ("wasm_singlethread", "wasm_multithread"):
        job_queue.append(
            BuildJob("install-qt", "6.5.0", host, "desktop", wasm_arch, wasm_arch,
                     is_autodesktop=True, emsdk_version="sdk-3.1.25-64bit", autodesk_arch_folder=desk_arch)
        )
mac_build_jobs.append(
    BuildJob("install-qt", "5.14.2", "mac", "desktop", "wasm_32", "wasm_32")
)
mac_build_jobs.append(
    BuildJob("install-qt", "6.4.0", "mac", "desktop", "wasm_32", "wasm_32",
             is_autodesktop=True, emsdk_version="sdk-3.1.14-64bit", autodesk_arch_folder="clang_64")
)
windows_build_jobs.append(
    BuildJob("install-qt", "5.14.2", "windows", "desktop", "wasm_32", "wasm_32")
)
windows_build_jobs.append(
    BuildJob("install-qt", "6.4.0", "windows", "desktop", "wasm_32", "wasm_32",
             is_autodesktop=True, emsdk_version="sdk-3.1.14-64bit", autodesk_arch_folder="mingw_64",
             mingw_variant="win64_mingw900")
)

# mobile SDK
mac_build_jobs.extend(
    [
        BuildJob("install-qt", "6.4.0", "mac", "ios", "ios", "ios", module="qtsensors", is_autodesktop=True),
        BuildJob("install-qt", "6.2.4", "mac", "ios", "ios", "ios", module="qtsensors", is_autodesktop=False),
        BuildJob("install-qt", "6.4.1", "mac", "android", "android_armv7", "android_armv7", is_autodesktop=True),
        BuildJob("install-qt", "6.1.0", "mac", "android", "android_armv7", "android_armv7", is_autodesktop=True),
    ]
)
linux_build_jobs.extend(
    [
        BuildJob("install-qt", "6.1.0", "linux", "android", "android_armv7", "android_armv7", is_autodesktop=True),
        BuildJob("install-qt", "6.4.1", "linux", "android", "android_arm64_v8a", "android_arm64_v8a", is_autodesktop=True),
    ]
)

# Qt 6.3.0 for Windows-Android has win64_mingw available, but not win64_mingw81.
# This will test that the path to mingw is not hardcoded.
windows_build_jobs.extend(
    [
        BuildJob("install-qt", "6.3.0", "windows", "android", "android_armv7", "android_armv7", is_autodesktop=True),
        BuildJob("install-qt", "6.4.1", "windows", "android", "android_x86_64", "android_x86_64", is_autodesktop=True),
    ]
)

# Test binary patch of qmake
linux_build_jobs.extend(
    [
        # New output dir is shorter than the default value; qmake could fail to
        # locate prefix dir if the value is patched wrong
        BuildJob(
            "install-qt",
            "5.12.11",
            "linux",
            "desktop",
            "gcc_64",
            "gcc_64",
            output_dir="/t/Q",
        ),
        # New output dir is longer than the default value.
        # This case is meant to work without any bugfix; if this fails, the test is setup wrong
        BuildJob(
            "install-qt",
            "5.12.11",
            "linux",
            "desktop",
            "gcc_64",
            "gcc_64",
            output_dir="/some/super/long/arbitrary/path/to" * 5,
        ),
    ]
)

qt_creator_bin_path = "./Tools/QtCreator/bin/"
qt_creator_mac_bin_path = "./Qt Creator.app/Contents/MacOS/"
qt_ifw_bin_path = "./Tools/QtInstallerFramework/*/bin/"
tool_options = {
    "TOOL1_ARGS": "tools_qtcreator qt.tools.qtcreator",
    "LIST_TOOL1_CMD": f"ls {qt_creator_bin_path}",
    "TEST_TOOL1_CMD": f"{qt_creator_bin_path}qbs --version",
    "TOOL2_ARGS": "tools_ifw",
    "TEST_TOOL2_CMD": f"{qt_ifw_bin_path}archivegen --version",
    "LIST_TOOL2_CMD": f"ls {qt_ifw_bin_path}",
}
# Mac Qt Creator is a .app, or "Package Bundle", so the path is changed:
tool_options_mac = {
    **tool_options,
    "TEST_TOOL1_CMD": f'"{qt_creator_mac_bin_path}qbs" --version',
    "LIST_TOOL1_CMD": f'ls "{qt_creator_mac_bin_path}"',
}
windows_build_jobs.append(
    BuildJob("install-tool", "", "windows", "desktop", "", "", tool_options=tool_options)
)
linux_build_jobs.append(
    BuildJob("install-tool", "", "linux", "desktop", "", "", tool_options=tool_options)
)
mac_build_jobs.append(
    BuildJob("install-tool", "", "mac", "desktop", "", "", tool_options=tool_options_mac)
)

matrices = {}

for platform_build_job in all_platform_build_jobs:
    matrix_dictionary = collections.OrderedDict()

    for build_job, python_version in product(
        platform_build_job.build_jobs, python_versions
    ):
        key = "{} {} {} for {}".format(
            build_job.command, build_job.qt_version, build_job.arch, build_job.target
        )
        if build_job.spec:
            key = '{} (spec="{}")'.format(key, build_job.spec)
        if build_job.module:
            key = "{} ({})".format(key, build_job.module)
        if build_job.subarchives:
            key = "{} ({})".format(key, build_job.subarchives)
        if build_job.output_dir:
            key = "{} ({})".format(key, build_job.output_dir)
        matrix_dictionary[key] = collections.OrderedDict(
            [
                ("PYTHON_VERSION", python_version),
                ("SUBCOMMAND", build_job.command),
                ("QT_VERSION", build_job.qt_version),
                ("HOST", build_job.host),
                ("TARGET", build_job.target),
                ("ARCH", build_job.arch),
                ("ARCHDIR", build_job.archdir),
                ("MODULE", build_job.module if build_job.module else ""),
                ("QT_BASE_MIRROR", build_job.mirror if build_job.mirror else ""),
                ("SUBARCHIVES", build_job.subarchives if build_job.subarchives else ""),
                ("SPEC", build_job.spec if build_job.spec else ""),
                ("MINGW_VARIANT", build_job.mingw_variant),
                ("MINGW_FOLDER", build_job.mingw_folder()),
                ("IS_AUTODESKTOP", str(build_job.is_autodesktop)),
                ("HAS_WASM", build_job.list_options.get("HAS_WASM", "True")),
                ("OUTPUT_DIR", build_job.output_dir if build_job.output_dir else ""),
                ("QT_BINDIR", build_job.qt_bindir()),
                ("WIN_QT_BINDIR", build_job.win_qt_bindir()),
                ("EMSDK_VERSION", (build_job.emsdk_version+"@main").split('@')[0]),
                ("EMSDK_TAG",  (build_job.emsdk_version+"@main").split('@')[1]),
                ("WIN_AUTODESK_QT_BINDIR", build_job.win_autodesk_qt_bindir()),
                ("TOOL1_ARGS", build_job.tool_options.get("TOOL1_ARGS", "")),
                ("LIST_TOOL1_CMD", build_job.tool_options.get("LIST_TOOL1_CMD", "")),
                ("TEST_TOOL1_CMD", build_job.tool_options.get("TEST_TOOL1_CMD", "")),
                ("TOOL2_ARGS", build_job.tool_options.get("TOOL2_ARGS", "")),
                ("LIST_TOOL2_CMD", build_job.tool_options.get("LIST_TOOL2_CMD", "")),
                ("TEST_TOOL2_CMD", build_job.tool_options.get("TEST_TOOL2_CMD", "")),
                ("CHECK_OUTPUT_CMD", build_job.check_output_cmd or "")
            ]
        )

    matrices[platform_build_job.platform] = matrix_dictionary

print("Setting Variables below")
print(
    f"##vso[task.setVariable variable=linux;isOutput=true]{json.dumps(matrices['linux'])}"
)
print(
    f"##vso[task.setVariable variable=windows;isOutput=true]{json.dumps(matrices['windows'])}"
)
print(
    f"##vso[task.setVariable variable=mac;isOutput=true]{json.dumps(matrices['mac'])}"
)



import re
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from aqt.archives import TargetConfig
from aqt.exceptions import UpdaterError
from aqt.helper import Settings
from aqt.updater import Updater


@pytest.fixture(autouse=True)
def setup_settings():
    Settings.load_settings()


@pytest.mark.parametrize(
    "target_config, expected_err_pattern",
    (
        (
            TargetConfig("5.15.2", "winrt", "win64_msvc2019_winrt_x86", "windows"),
            re.compile(
                r"Updater caused an IO error: .*No such file or directory: "
                # '.*' wildcard used to match path separators on windows/*nix
                r".*5\.15\.2.*winrt_x86_msvc2019.*mkspecs.*qconfig.pri.*"
            ),
        ),
        (
            TargetConfig("5.15.2", "desktop", "win64_msvc2019_64", "windows"),
            re.compile(
                r"Updater caused an IO error: .*No such file or directory: "
                # '.*' wildcard used to match path separators on windows/*nix
                r".*5\.15\.2.*msvc2019_64.*mkspecs.*qconfig.pri.*"
            ),
        ),
        (
            TargetConfig("6.1.2", "desktop", "clang_64", "mac"),
            re.compile(
                r"Updater caused an IO error: .*No such file or directory: "
                # '.*' wildcard used to match path separators on windows/*nix
                r".*6\.1\.2.*macos.*mkspecs.*qconfig.pri.*"
            ),
        ),
        (
            TargetConfig("6.1.1", "desktop", "clang_64", "mac"),
            re.compile(
                r"Updater caused an IO error: .*No such file or directory: "
                # '.*' wildcard used to match path separators on windows/*nix
                r".*6\.1\.1.*clang_64.*mkspecs.*qconfig.pri.*"
            ),
        ),
    ),
)
def test_updater_update_license_io_error(monkeypatch, target_config: TargetConfig, expected_err_pattern: re.Pattern):
    """
    All of these tests will raise IOError when they attempt to patch the license file.
    """
    with pytest.raises(UpdaterError) as err:
        with TemporaryDirectory() as empty_dir:
            # Try to update a Qt installation that does not exist
            Updater.update(target_config, base_path=Path(empty_dir), installed_desktop_arch_dir=None)
    assert err.type == UpdaterError
    err_msg = format(err.value)
    assert expected_err_pattern.match(err_msg)
